#  Copyright (c) 2023. Davi Pereira dos Santos
#  This file is part of the hdict project.
#  Please respect the license - more about this in the section (*) below.
#
#  hdict is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  hdict is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with hdict.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is illegal and it is unethical regarding the effort and
#  time spent here.
#
from typing import TypeVar, Union

from hdict.data.frozenhdict import frozenhdict

VT = TypeVar("VT")


# todo: : finish all '*' combinations and check all '>>' to see when to generate hdict and when to generate Expr.


class hdict_(dict[str, VT]):
    """
    This class was created only due to slowness of IDE created by excessive doctests in the main file.

    >>> from hdict import _, apply
    >>> e = hdict_() * dict() * hdict_()
    >>> e.show(colored=False)
    ⦑{
        _id: 0000000000000000000000000000000000000000,
        _ids: {}
    } » {} » {}⦒
    >>> d = {"x": 3, "f": lambda x: x+1} >> e >> apply(_.f, _.x).x
    >>> bool(d), bool(e.sample())
    (True, False)
    >>> d.show(colored=False)  # doctest:+ELLIPSIS
    {
        x: λ(3),
        f: "<function <lambda> at 0x...>",
        _id: KxAogdPlPbTXATJzW7E8C.4j.82ZIDrVpTbVRxiZ,
        _ids: {
            x: KeU-dCTjUgnSYGRNrrMMr4i.hg64Dkkfb3c14eh3,
            f: p-nM7oHlOFr6iHSF9ZISWXc9zqi017c3zcvcV8Dr
        }
    }
    >>> d.x
    4
    >>> d.show(colored=False)  # doctest:+ELLIPSIS
    {
        x: 4,
        f: "<function <lambda> at 0x...>",
        _id: KxAogdPlPbTXATJzW7E8C.4j.82ZIDrVpTbVRxiZ,
        _ids: {
            x: KeU-dCTjUgnSYGRNrrMMr4i.hg64Dkkfb3c14eh3,
            f: p-nM7oHlOFr6iHSF9ZISWXc9zqi017c3zcvcV8Dr
        }
    }
    >>> str({"a": 2} * d)  # doctest:+ELLIPSIS
    '⦑{a: 2} » {x: 4, f: "<function <lambda> at 0x...>"}⦒'
    >>> (d >> apply(lambda d: d.x, _).X).evaluated.show(colored=False)  # doctest:+ELLIPSIS
    {
        x: 4,
        f: "<function <lambda> at 0x...>",
        X: 4,
        _id: KaIR9dqecR44MZVRrXLUzwqnIbCHhbOkQswgCuQl,
        _ids: {
            x: KeU-dCTjUgnSYGRNrrMMr4i.hg64Dkkfb3c14eh3,
            f: p-nM7oHlOFr6iHSF9ZISWXc9zqi017c3zcvcV8Dr,
            X: M411IcX.kF9pEdsWGIiS-XGn.R45qYTVQS0ewQs2
        }
    }
    >>> (d >> apply(lambda _: _.x).X).evaluated.show(colored=False)  # doctest:+ELLIPSIS
    {
        x: 4,
        f: "<function <lambda> at 0x...>",
        X: 4,
        _id: c1ZgriP45q8xEmDTkRcHjYe2v1C7RLyyZK4NgQ0L,
        _ids: {
            x: KeU-dCTjUgnSYGRNrrMMr4i.hg64Dkkfb3c14eh3,
            f: p-nM7oHlOFr6iHSF9ZISWXc9zqi017c3zcvcV8Dr,
            X: nVVYNVw8RyQ-7Kv0CxAUeLjRIzZwZwE7-8BKaaFr
        }
    }
    """

    # noinspection PyMissingConstructor
    def __init__(self, /, _dictionary: dict = None, _frozen: frozenhdict = None, **kwargs):
        # Build hdict from frozen or fresh data. Never both.
        self.frozen = _frozen or frozenhdict(_dictionary, **kwargs)
        self.raw = self.frozen.data

    def __setitem__(self, key: str | tuple, value):
        if isinstance(key, tuple):
            key = tuple((x.start, x.stop) if isinstance(x, slice) else x for x in key)
        self.frozen = self.frozen >> {key: value}

    def __delitem__(self, key):
        data = self.frozen.data.copy()
        del data[key]
        self.frozen = frozenhdict(data)

    def __getitem__(self, item):
        return self.frozen[item]

    def __getattr__(self, item):
        if item in self.frozen:
            return self.frozen[item]
        return self.__getattribute__(item)  # pragma: no cover

    def __rmul__(self, other):
        return other * self.frozen

    def __mul__(self, other):
        return self.frozen * other

    def __rrshift__(self, other):
        return (other >> self.frozen).unfrozen

    def __rshift__(self, other):
        return (self.frozen >> other).unfrozen

    @property
    def evaluated(self):
        return self.frozen.evaluated

    def evaluate(self):
        """
        >>> from hdict import apply, hdict
        >>> d = hdict(x=apply(apply(lambda: 2)))
        >>> d.show(colored=False)
        {
            x: λ(),
            _id: YwRX7paX43aafhz-Jndo9HYnvyMSOtPEbRDjgQ3r,
            _ids: {
                x: J746LLRT3gd5glC2ZiaHBgHegmewpiOMC4Sq4bp9
            }
        }
        >>> d.evaluate()
        >>> d.show(colored=False)
        {
            x: 2,
            _id: YwRX7paX43aafhz-Jndo9HYnvyMSOtPEbRDjgQ3r,
            _ids: {
                x: J746LLRT3gd5glC2ZiaHBgHegmewpiOMC4Sq4bp9
            }
        }
        >>> d.evaluated.show(colored=False)
        {
            x: 2,
            _id: YwRX7paX43aafhz-Jndo9HYnvyMSOtPEbRDjgQ3r,
            _ids: {
                x: J746LLRT3gd5glC2ZiaHBgHegmewpiOMC4Sq4bp9
            }
        }
        """
        self.frozen.evaluate()

    @property
    def hosh(self):
        return self.frozen.hosh

    @property
    def id(self):
        """
        >>> from hdict import hdict
        >>> hdict(x=3, y=5).id == hdict(dict(x=3, y=5)).id
        True
        """
        return self.hosh.id

    @property
    def ids(self):
        """
        >>> from hdict import hdict
        >>> hdict(x=3, y=5).ids
        {'x': 'KGWjj0iyLAn1RG6RTGtsGE3omZraJM6xO.kvG5pr', 'y': 'ecvgo-CBPi7wRWIxNzuo1HgHQCbdvR058xi6zmr2'}
        """
        return self.frozen.ids

    @staticmethod
    def fromdict(dictionary, ids):
        """Build an idict from values and pre-defined ids

        >>> from hosh import Hosh
        >>> from hdict import hdict
        >>> e = hdict.fromdict({"x": 3, "y": 5, "z": 7}, ids={"x": Hosh(b"x"), "y": Hosh(b"y").id})
        >>> e.show(colored=False)
        {
            x: 3,
            y: 5,
            z: 7,
            _id: uf--zyyiojm5Tl.vFKALuyGhZRO0e0eH9irosr0i,
            _ids: {
                x: ue7X2I7fd9j0mLl1GjgJ2btdX1QFnb1UAQNUbFGh,
                y: 5yg5fDxFPxhEqzhoHgXpKyl5f078iBhd.pR0G2X0,
                z: eJCW9jGsdZTD6-AD9opKwjPIOWZ4R.T0CG2kdyzf
            }
        }
        """
        from hdict.data.frozenhdict import frozenhdict

        return frozenhdict.fromdict(dictionary, ids).unfrozen

    @property
    def asdict(self):
        """
        Convert to 'dict', including ids.

        This evaluates all fields.
        HINT: Use 'dict(d)' to convert 'hdict' to 'dict' excluding ids.

        >>> from hdict import hdict
        >>> hdict(x=3, y=5).asdict
        {'x': 3, 'y': 5, '_id': 'r5A2Mh6vRRO5rxi5nfXv1myeguGSTmqHuHev38qM', '_ids': {'x': 'KGWjj0iyLAn1RG6RTGtsGE3omZraJM6xO.kvG5pr', 'y': 'ecvgo-CBPi7wRWIxNzuo1HgHQCbdvR058xi6zmr2'}}
        """
        return self.frozen.asdict

    @property
    def asdicts(self):
        """
        >>> from hdict import hdict
        >>> hdict(x=3, y=5).asdict
        {'x': 3, 'y': 5, '_id': 'r5A2Mh6vRRO5rxi5nfXv1myeguGSTmqHuHev38qM', '_ids': {'x': 'KGWjj0iyLAn1RG6RTGtsGE3omZraJM6xO.kvG5pr', 'y': 'ecvgo-CBPi7wRWIxNzuo1HgHQCbdvR058xi6zmr2'}}
        """
        return self.frozen.asdicts

    @property
    def asdicts_noid(self):
        return self.frozen.asdicts_noid

    def astext(self, colored=True, key_quotes=False):
        r"""
        >>> from hdict import hdict
        >>> repr(hdict(x=3, y=5)) == hdict(x=3, y=5).astext()
        True
        >>> print(repr(hdict(x=3, y=5)))
        {
            x: 3,
            y: 5,
            _id: [38;5;225m[1m[48;5;0mr[0m[38;5;225m[1m[48;5;0m5[0m[38;5;15m[1m[48;5;0mA[0m[38;5;225m[1m[48;5;0m2[0m[38;5;225m[1m[48;5;0mM[0m[38;5;195m[1m[48;5;0mh[0m[38;5;225m[1m[48;5;0m6[0m[38;5;219m[1m[48;5;0mv[0m[38;5;183m[1m[48;5;0mR[0m[38;5;251m[1m[48;5;0mR[0m[38;5;177m[1m[48;5;0mO[0m[38;5;147m[1m[48;5;0m5[0m[38;5;183m[1m[48;5;0mr[0m[38;5;189m[1m[48;5;0mx[0m[38;5;15m[1m[48;5;0mi[0m[38;5;225m[1m[48;5;0m5[0m[38;5;225m[1m[48;5;0mn[0m[38;5;225m[1m[48;5;0mf[0m[38;5;15m[1m[48;5;0mX[0m[38;5;225m[1m[48;5;0mv[0m[38;5;225m[1m[48;5;0m1[0m[38;5;195m[1m[48;5;0mm[0m[38;5;225m[1m[48;5;0my[0m[38;5;219m[1m[48;5;0me[0m[38;5;183m[1m[48;5;0mg[0m[38;5;251m[1m[48;5;0mu[0m[38;5;177m[1m[48;5;0mG[0m[38;5;147m[1m[48;5;0mS[0m[38;5;183m[1m[48;5;0mT[0m[38;5;189m[1m[48;5;0mm[0m[38;5;15m[1m[48;5;0mq[0m[38;5;225m[1m[48;5;0mH[0m[38;5;225m[1m[48;5;0mu[0m[38;5;225m[1m[48;5;0mH[0m[38;5;15m[1m[48;5;0me[0m[38;5;225m[1m[48;5;0mv[0m[38;5;225m[1m[48;5;0m3[0m[38;5;195m[1m[48;5;0m8[0m[38;5;225m[1m[48;5;0mq[0m[38;5;219m[1m[48;5;0mM[0m,
            _ids: {
                x: [38;5;239m[1m[48;5;0mK[0m[38;5;239m[1m[48;5;0mG[0m[38;5;60m[1m[48;5;0mW[0m[38;5;241m[1m[48;5;0mj[0m[38;5;97m[1m[48;5;0mj[0m[38;5;65m[1m[48;5;0m0[0m[38;5;133m[1m[48;5;0mi[0m[38;5;65m[1m[48;5;0my[0m[38;5;97m[1m[48;5;0mL[0m[38;5;66m[1m[48;5;0mA[0m[38;5;132m[1m[48;5;0mn[0m[38;5;61m[1m[48;5;0m1[0m[38;5;66m[1m[48;5;0mR[0m[38;5;95m[1m[48;5;0mG[0m[38;5;95m[1m[48;5;0m6[0m[38;5;239m[1m[48;5;0mR[0m[38;5;239m[1m[48;5;0mT[0m[38;5;239m[1m[48;5;0mG[0m[38;5;60m[1m[48;5;0mt[0m[38;5;241m[1m[48;5;0ms[0m[38;5;97m[1m[48;5;0mG[0m[38;5;65m[1m[48;5;0mE[0m[38;5;133m[1m[48;5;0m3[0m[38;5;65m[1m[48;5;0mo[0m[38;5;97m[1m[48;5;0mm[0m[38;5;66m[1m[48;5;0mZ[0m[38;5;132m[1m[48;5;0mr[0m[38;5;61m[1m[48;5;0ma[0m[38;5;66m[1m[48;5;0mJ[0m[38;5;95m[1m[48;5;0mM[0m[38;5;95m[1m[48;5;0m6[0m[38;5;239m[1m[48;5;0mx[0m[38;5;239m[1m[48;5;0mO[0m[38;5;239m[1m[48;5;0m.[0m[38;5;60m[1m[48;5;0mk[0m[38;5;241m[1m[48;5;0mv[0m[38;5;97m[1m[48;5;0mG[0m[38;5;65m[1m[48;5;0m5[0m[38;5;133m[1m[48;5;0mp[0m[38;5;65m[1m[48;5;0mr[0m,
                y: [38;5;227m[1m[48;5;0me[0m[38;5;221m[1m[48;5;0mc[0m[38;5;209m[1m[48;5;0mv[0m[38;5;149m[1m[48;5;0mg[0m[38;5;221m[1m[48;5;0mo[0m[38;5;215m[1m[48;5;0m-[0m[38;5;185m[1m[48;5;0mC[0m[38;5;221m[1m[48;5;0mB[0m[38;5;216m[1m[48;5;0mP[0m[38;5;192m[1m[48;5;0mi[0m[38;5;227m[1m[48;5;0m7[0m[38;5;222m[1m[48;5;0mw[0m[38;5;191m[1m[48;5;0mR[0m[38;5;215m[1m[48;5;0mW[0m[38;5;180m[1m[48;5;0mI[0m[38;5;192m[1m[48;5;0mx[0m[38;5;227m[1m[48;5;0mN[0m[38;5;221m[1m[48;5;0mz[0m[38;5;209m[1m[48;5;0mu[0m[38;5;149m[1m[48;5;0mo[0m[38;5;221m[1m[48;5;0m1[0m[38;5;215m[1m[48;5;0mH[0m[38;5;185m[1m[48;5;0mg[0m[38;5;221m[1m[48;5;0mH[0m[38;5;216m[1m[48;5;0mQ[0m[38;5;192m[1m[48;5;0mC[0m[38;5;227m[1m[48;5;0mb[0m[38;5;222m[1m[48;5;0md[0m[38;5;191m[1m[48;5;0mv[0m[38;5;215m[1m[48;5;0mR[0m[38;5;180m[1m[48;5;0m0[0m[38;5;192m[1m[48;5;0m5[0m[38;5;227m[1m[48;5;0m8[0m[38;5;221m[1m[48;5;0mx[0m[38;5;209m[1m[48;5;0mi[0m[38;5;149m[1m[48;5;0m6[0m[38;5;221m[1m[48;5;0mz[0m[38;5;215m[1m[48;5;0mm[0m[38;5;185m[1m[48;5;0mr[0m[38;5;221m[1m[48;5;0m2[0m
            }
        }
        """
        return self.frozen.astext(colored, key_quotes)

    def show(self, colored=True, key_quotes=False):
        r"""
        Textual representation of an idict object

        >>> from hdict import hdict
        >>> d = hdict(x=1,y=2)
        >>> d.show(colored=False)
        {
            x: 1,
            y: 2,
            _id: 41wHsGFMSo0vbD2n6zAXogYG9YE3FwzIRSqjWc8N,
            _ids: {
                x: DYu5bfVvb6FOhBCWNsss4wsEWHZYTbKnsVgoWFvl,
                y: k3PWYRxIEc0lEvD1f6rbnk.36RAD5AyfROy1aT29
            }
        }
        """
        return self.frozen.show(colored, key_quotes)

    def __iter__(self):
        return iter(self.frozen)

    def __contains__(self, item):
        return item in self.frozen

    def __repr__(self):
        return repr(self.frozen)

    def __str__(self):
        return str(self.frozen)

    def __eq__(self, other):
        return self.frozen == other

    def __ne__(self, other):
        return self.frozen != other

    def keys(self):
        """Generator of field names, i.e., keys which don't start with '_'"""
        return self.frozen.keys()

    def values(self, evaluate=True):
        """Generator of field values (keys that don't start with '_')"""
        return self.frozen.values(evaluate)

    def items(self, evaluate=True):
        """Generator over field-value pairs

        Exclude ids and other items starting with '_'.

        >>> from hdict import hdict
        >>> for k, v in hdict(x=1, y=2).items():
        ...     print(k, v)
        x 1
        y 2
        """
        return self.frozen.items(evaluate)

    def __hash__(self):  # pragma: no cover
        raise Exception(f"hdict is not hashable. Please use hdict.frozen instead.")

    def save(self, storage: dict):
        """
        Store an entire hdict

        >>> from hdict import hdict, apply
        >>> storage = {}
        >>> d = hdict(x=3, y=7, z=hdict(z=9)) >> apply(lambda x, y: x/y).w
        >>> d.show(colored=False)
        {
            x: 3,
            y: 7,
            z: {
                z: 9,
                _id: izn67XbX0tQNF6E5qkwniN2jxZg5MT6f7z5AJzPM,
                _ids: {
                    z: GuwIQCrendfKXZr5jGfrUwoP-8TWMhmLHYrja2yj
                }
            },
            w: λ(x y),
            _id: Jf8QtLAszezNdiqblJCXMuoXocCWZ90QwrxY90Zu,
            _ids: {
                x: KGWjj0iyLAn1RG6RTGtsGE3omZraJM6xO.kvG5pr,
                y: eJCW9jGsdZTD6-AD9opKwjPIOWZ4R.T0CG2kdyzf,
                z: izn67XbX0tQNF6E5qkwniN2jxZg5MT6f7z5AJzPM,
                w: NjWbIhEnN3s8E-w61pptxBuiYmqQ2D3QNX2ZAEB1
            }
        }
        >>> d.save(storage)
        >>> storage
        {'izn67XbX0tQNF6E5qkwniN2jxZg5MT6f7z5AJzPM': {'z': 'GuwIQCrendfKXZr5jGfrUwoP-8TWMhmLHYrja2yj'}, 'GuwIQCrendfKXZr5jGfrUwoP-8TWMhmLHYrja2yj': Stored(content=9), 'Jf8QtLAszezNdiqblJCXMuoXocCWZ90QwrxY90Zu': {'x': 'KGWjj0iyLAn1RG6RTGtsGE3omZraJM6xO.kvG5pr', 'y': 'eJCW9jGsdZTD6-AD9opKwjPIOWZ4R.T0CG2kdyzf', 'z': 'izn67XbX0tQNF6E5qkwniN2jxZg5MT6f7z5AJzPM', 'w': 'NjWbIhEnN3s8E-w61pptxBuiYmqQ2D3QNX2ZAEB1'}, 'KGWjj0iyLAn1RG6RTGtsGE3omZraJM6xO.kvG5pr': Stored(content=3), 'eJCW9jGsdZTD6-AD9opKwjPIOWZ4R.T0CG2kdyzf': Stored(content=7), 'NjWbIhEnN3s8E-w61pptxBuiYmqQ2D3QNX2ZAEB1': Stored(content=0.42857142857142855)}
        >>> e = hdict.load(d.id, storage)
        >>> e.show(colored=False)
        {
            x: ↑↓ cached at `dict`·,
            y: ↑↓ cached at `dict`·,
            z: ↑↓ cached at `dict`·,
            w: ↑↓ cached at `dict`·,
            _id: Jf8QtLAszezNdiqblJCXMuoXocCWZ90QwrxY90Zu,
            _ids: {
                x: KGWjj0iyLAn1RG6RTGtsGE3omZraJM6xO.kvG5pr,
                y: eJCW9jGsdZTD6-AD9opKwjPIOWZ4R.T0CG2kdyzf,
                z: izn67XbX0tQNF6E5qkwniN2jxZg5MT6f7z5AJzPM,
                w: NjWbIhEnN3s8E-w61pptxBuiYmqQ2D3QNX2ZAEB1
            }
        }
        >>> d.w
        0.42857142857142855
        >>> e.w
        0.42857142857142855
        >>> e.evaluate()
        >>> e.show(colored=False)
        {
            x: 3,
            y: 7,
            z: {
                z: 9
            },
            w: 0.42857142857142855,
            _id: Jf8QtLAszezNdiqblJCXMuoXocCWZ90QwrxY90Zu,
            _ids: {
                x: KGWjj0iyLAn1RG6RTGtsGE3omZraJM6xO.kvG5pr,
                y: eJCW9jGsdZTD6-AD9opKwjPIOWZ4R.T0CG2kdyzf,
                z: izn67XbX0tQNF6E5qkwniN2jxZg5MT6f7z5AJzPM,
                w: NjWbIhEnN3s8E-w61pptxBuiYmqQ2D3QNX2ZAEB1
            }
        }
        """
        self.frozen.save(storage)

    @staticmethod
    def fetch(id: str, storage: dict, lazy=True, ishdict=False) -> Union["hdict_", None]:
        r"""
        Fetch a single entry

        When cache is a list, traverse it from the end (right item to the left item).

        >>> from hdict import hdict, cache
        >>> from testfixtures import TempDirectory
        >>> arff = "@RELATION mini\n@ATTRIBUTE attr1	REAL\n@ATTRIBUTE attr2 	REAL\n@ATTRIBUTE class 	{0,1}\n@DATA\n5.1,3.5,0\n3.1,4.5,1"
        >>> with TempDirectory() as tmp:  # doctest:+ELLIPSIS
        ...    tmp.write("mini.arff", arff.encode())
        ...    d = hdict.fromfile(tmp.path + "/mini.arff", fields=["df"])
        ...    d2 = hdict.fromfile(tmp.path + "/mini.arff", fields=["df"], named=True)
        '/tmp/.../mini.arff'
        >>> d.show(colored=False)
        {
            df: "‹{'attr1': {0: 5.1, 1: 3.1}, 'attr2': {0: 3.5, 1: 4.5}, 'class': {0: '0', 1: '1'}}›",
            _id: YaEOJ7a7aOUux8qb1lY3mucUo6AU4vVv15JhAnn2,
            _ids: {
                df: Sxa-cdrwGU60j--6mK9FvWc7IaYdeYB-2yzSi62B
            }
        }
        >>> d2.show(colored=False)
        {
            df: "‹{'attr1': {0: 5.1, 1: 3.1}, 'attr2': {0: 3.5, 1: 4.5}, 'class': {0: '0', 1: '1'}}›",
            name: " mini",
            _id: NuQ3MtEDc-RqAgxUtgEJdg1wnQngv2koODT8GrqU,
            _ids: {
                df: Sxa-cdrwGU60j--6mK9FvWc7IaYdeYB-2yzSi62B,
                name: 5nzeWHgeVef.PD8CnIrPcjwGqaYoRCXK0iwH0cKm
            }
        }
        >>> storage = {}
        >>> d.save(storage)
        >>> d = hdict.fetch(d.id, storage)
        >>> d.show(colored=False)
        {
            df: ↑↓ cached at `dict`·,
            _id: YaEOJ7a7aOUux8qb1lY3mucUo6AU4vVv15JhAnn2,
            _ids: {
                df: Sxa-cdrwGU60j--6mK9FvWc7IaYdeYB-2yzSi62B
            }
        }
        >>> d.df
           attr1  attr2 class
        0    5.1    3.5     0
        1    3.1    4.5     1

        # >>> #d.df.show(colored=False)
        # >>> d = hdict.fetch(d.id, storage, lazy=False)
        # >>> d.show(colored=False)
        # {
        #     df: {
        #         index: "‹{0: 0, 1: 1}›",
        #         "attr1@REAL": "‹{0: 5.1, 1: 3.1}›",
        #         "attr2@REAL": "‹{0: 3.5, 1: 4.5}›",
        #         "class@{0,1}": "‹{0: '0', 1: '1'}›",
        #         _id: cHrG-npBDd2VEB8Foeg.7jQNZtdkTM1uhouHgW.J,
        #         _ids: {
        #             index: DQa5yWRkGo-9FLqmaST8pbElYdUEgqF8xPvip6-3,
        #             "attr1@REAL": wsy0JDrZO04O0RVwr64jpawX62WmCKtddYdvZlwm,
        #             "attr2@REAL": LvEgUazJoB1-.A3kABbskN-.iauWmWLTTo1iC51n,
        #             "class@{0,1}": b.kJy3SrU-JQ1oeh1d.uWLO7Pqh-eW6zwK78nTY4
        #         }
        #     },
        #     df_: λ({    i···),
        #     _id: CWkreYbSmrL0DPN9OtoU4Za1dg8.Jjl.fXD6yblb,
        #     _ids: {
        #         df: cHrG-npBDd2VEB8Foeg.7jQNZtdkTM1uhouHgW.J,
        #         df_: cHrG-npBDd2VEB8Foeg.7jQNZtdkTM1uhouHgW.J
        #     }
        # }
        # >>> #del d["df_"]
        # >>> d.show(colored=False)
        # {
        #     df: {
        #         index: "‹{0: 0, 1: 1}›",
        #         "attr1@REAL": "‹{0: 5.1, 1: 3.1}›",
        #         "attr2@REAL": "‹{0: 3.5, 1: 4.5}›",
        #         "class@{0,1}": "‹{0: '0', 1: '1'}›",
        #         _id: cHrG-npBDd2VEB8Foeg.7jQNZtdkTM1uhouHgW.J,
        #         _ids: {
        #             index: DQa5yWRkGo-9FLqmaST8pbElYdUEgqF8xPvip6-3,
        #             "attr1@REAL": wsy0JDrZO04O0RVwr64jpawX62WmCKtddYdvZlwm,
        #             "attr2@REAL": LvEgUazJoB1-.A3kABbskN-.iauWmWLTTo1iC51n,
        #             "class@{0,1}": b.kJy3SrU-JQ1oeh1d.uWLO7Pqh-eW6zwK78nTY4
        #         }
        #     },
        #     _id: CWkreYbSmrL0DPN9OtoU4Za1dg8.Jjl.fXD6yblb,
        #     _ids: {
        #         df: cHrG-npBDd2VEB8Foeg.7jQNZtdkTM1uhouHgW.J
        #     }
        # }
        # >>> d.save(storage)
        # >>> d = hdict.fetch(d.id, storage, lazy=False)
        # >>> d.show(colored=False)
        # {
        #     df: {
        #         index: "‹{0: 0, 1: 1}›",
        #         "attr1@REAL": "‹{0: 5.1, 1: 3.1}›",
        #         "attr2@REAL": "‹{0: 3.5, 1: 4.5}›",
        #         "class@{0,1}": "‹{0: '0', 1: '1'}›",
        #         _id: cHrG-npBDd2VEB8Foeg.7jQNZtdkTM1uhouHgW.J,
        #         _ids: {
        #             index: DQa5yWRkGo-9FLqmaST8pbElYdUEgqF8xPvip6-3,
        #             "attr1@REAL": wsy0JDrZO04O0RVwr64jpawX62WmCKtddYdvZlwm,
        #             "attr2@REAL": LvEgUazJoB1-.A3kABbskN-.iauWmWLTTo1iC51n,
        #             "class@{0,1}": b.kJy3SrU-JQ1oeh1d.uWLO7Pqh-eW6zwK78nTY4
        #         }
        #     },
        #     _id: CWkreYbSmrL0DPN9OtoU4Za1dg8.Jjl.fXD6yblb,
        #     _ids: {
        #         df: cHrG-npBDd2VEB8Foeg.7jQNZtdkTM1uhouHgW.J
        #     }
        # }
        # >>> str(d)
        # '{df: {index: "‹{0: 0, 1: 1}›", attr1@REAL: "‹{0: 5.1, 1: 3.1}›", attr2@REAL: "‹{0: 3.5, 1: 4.5}›", class@{0,1}: "‹{0: \'0\', 1: \'1\'}›"}}'
        """
        return frozenhdict.fetch(id, storage, lazy, ishdict).unfrozen

    @staticmethod
    def load(id, storage: dict):
        """
        Fetch an entire hdict

        >>> from hdict import _
        >>> from hdict.persistence.stored import Stored
        >>> fid = "1234567890123456789012345678901234567890"
        >>> did = "0000567890123456789012345678901234567890"
        >>> storage = {did: {"x": fid}, fid: Stored(5)}
        >>> d = (+_).load(did, storage)
        >>> d.show(colored=False)
        {
            x: ↑↓ cached at `dict`·,
            _id: kYzgpPdRgQSYSEpp1qt4EHQLQJXuyb2WDQS-iNPh,
            _ids: {
                x: 1234567890123456789012345678901234567890
            }
        }
        >>> d.evaluate()
        >>> d.show(colored=False)
        {
            x: 5,
            _id: kYzgpPdRgQSYSEpp1qt4EHQLQJXuyb2WDQS-iNPh,
            _ids: {
                x: 1234567890123456789012345678901234567890
            }
        }
        """
        return frozenhdict.load(id, storage).unfrozen

    @staticmethod
    def fromfile(name, fields=None, format="df", named=None, hide_types=True):
        r"""
        Input format is defined by file extension: .arff, .csv

        >>> from hdict import hdict
        >>> from testfixtures import TempDirectory
        >>> arff = "@RELATION mini\n@ATTRIBUTE attr1	REAL\n@ATTRIBUTE attr2 	REAL\n@ATTRIBUTE class 	{0,1}\n@DATA\n5.1,3.5,0\n3.1,4.5,1"
        >>> with TempDirectory() as tmp:  # doctest:+ELLIPSIS
        ...    tmp.write("mini.arff", arff.encode())
        ...    d = hdict.fromfile(tmp.path + "/mini.arff")
        '/tmp/.../mini.arff'
        >>> d.show(colored=False)
        {
            df: "‹{'attr1': {0: 5.1, 1: 3.1}, 'attr2': {0: 3.5, 1: 4.5}, 'class': {0: '0', 1: '1'}}›",
            _id: YaEOJ7a7aOUux8qb1lY3mucUo6AU4vVv15JhAnn2,
            _ids: {
                df: Sxa-cdrwGU60j--6mK9FvWc7IaYdeYB-2yzSi62B
            }
        }
        >>> csv = "attr1,attr2,class\n5.1,3.5,0\n3.1,4.5,1"
        >>> with TempDirectory() as tmp:  # doctest:+ELLIPSIS
        ...    tmp.write("mini.csv", csv.encode())
        ...    d = hdict.fromfile(tmp.path + "/mini.csv")
        '/tmp/.../mini.csv'
        >>> d.show(colored=False)
        {
            df: "‹{'attr1': {0: 5.1, 1: 3.1}, 'attr2': {0: 3.5, 1: 4.5}, 'class': {0: 0, 1: 1}}›",
            _id: NPf36ZQWhyTj8yV3hVNjlKN3qJTNIxjgDJAkMWJY,
            _ids: {
                df: ytTKB-58EG6oWjgbY2-a-xyZJhBWUT.KESXVuFov
            }
        }
        """
        return frozenhdict.fromfile(name, fields, format, named, hide_types).unfrozen

    @staticmethod
    def fromtext(text: str, fields=None, format="df", named=None):
        r"""
        Input format is defined by file extension: .arff, .csv

        >>> from hdict import hdict
        >>> arff = "@RELATION mini\n@ATTRIBUTE attr1	REAL\n@ATTRIBUTE attr2 	REAL\n@ATTRIBUTE class 	{0,1}\n@DATA\n5.1,3.5,0\n3.1,4.5,1"
        >>> d = hdict.fromtext(arff)
        >>> d.show(colored=False)
        {
            df: "‹{'attr1@REAL': {0: 5.1, 1: 3.1}, 'attr2@REAL': {0: 3.5, 1: 4.5}, 'class@{0,1}': {0: '0', 1: '1'}}›",
            _id: zkG1Kyr0K0matB3G0-BvKmIi2e.dHhA4zA6X6DwO,
            _ids: {
                df: FPE1eyMzyycW6Ue9SqyxaiakR3H3VDgzAFtwRlbl
            }
        }
        >>> csv = "attr1,attr2,class\n5.1,3.5,0\n3.1,4.5,1"
        >>> d = hdict.fromtext(csv)
        >>> d.show(colored=False)
        {
            df: "‹{'attr1': {0: 5.1, 1: 3.1}, 'attr2': {0: 3.5, 1: 4.5}, 'class': {0: 0, 1: 1}}›",
            _id: NPf36ZQWhyTj8yV3hVNjlKN3qJTNIxjgDJAkMWJY,
            _ids: {
                df: ytTKB-58EG6oWjgbY2-a-xyZJhBWUT.KESXVuFov
            }
        }
        """
        return frozenhdict.fromtext(text, fields, format, named).unfrozen

    @property
    def asdf(self):
        """
        Represent hdict as a DataFrame if possible

        >>> from hdict import hdict
        >>> d = hdict({"x": [1,2,3], "y": [5,6,7], "index": ["a", "b", "c"]})
        >>> d.asdf
           x  y
        a  1  5
        b  2  6
        c  3  7
        """
        return self.frozen.asdf

    @property
    def hoshes(self):
        """
        >>> from hdict import frozenhdict
        >>> [h.id for h in frozenhdict(x=3, y=2).hoshes.values()]
        ['KGWjj0iyLAn1RG6RTGtsGE3omZraJM6xO.kvG5pr', 'k3PWYRxIEc0lEvD1f6rbnk.36RAD5AyfROy1aT29']
        """
        return self.frozen.hoshes

    def __bool__(self):
        return bool(self.frozen.data)

    # def __reduce__(self):
    # return self.frozen.__reduce__()



#  Copyright (c) 2023. Davi Pereira dos Santos
#  This file is part of the hdict project.
#  Please respect the license - more about this in the section (*) below.
#
#  hdict is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  hdict is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with hdict.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is illegal and it is unethical regarding the effort and
#  time spent here.
#

from dataclasses import dataclass
from random import Random

from hdict.content.argument.apply import apply
from hdict.expression.step.step import AbsStep


@dataclass
class ApplyOut(AbsStep):
    """
    Wrapper for 'apply' to append the output field(s)

    >>> from hdict import cache, apply, sample
    >>> e = dict() * apply(lambda x: x*2, x=sample(1,2,3,...,9)).a * (cache({}) >> dict(s=sample(1,2,3,...,9)))
    >>> e.show(colored=False)
    ⦑{} » a=λ(x=~[1 2 .+. 9]) » ↑↓`dict` » {s: "~[1 2 .+. 9]"}⦒
    >>> e.steps[0].steps[1].sampleable
    True
    >>> (~e).show(colored=False)
    {
        a: ↑↓ cached at `dict`·,
        s: 7,
        _id: E3YKpiQXWwK1AZh9BcqI7AoamBanxAr1GAOeJudy,
        _ids: {
            a: ygBt7xenihJVMnjwJxN6Z1H7HYnrET04zEYTRKD3,
            s: eJCW9jGsdZTD6-AD9opKwjPIOWZ4R.T0CG2kdyzf
        }
    }
    """

    nested: apply
    out: [str | tuple[str, str]]

    def sample(self, rnd: int | Random = None):
        if not self.nested.sampleable:
            return self
        return ApplyOut(self.nested.sample(rnd), self.out)

    @property
    def sampleable(self):
        return self.nested.sampleable

    def __repr__(self):
        return f"{self.out}={repr(self.nested)}"



from random import Random

from hdict.abs import AbsAny


class AbsArgument(AbsAny):
    """
    Argument provided by the user alongside the applied function

    For classes where sampling makes no sense, return the object itself.
    """

    sampleable = False

    def sample(self, rnd: int | Random = None):
        return self


class AbsBaseArgument(AbsArgument):
    """
    Normal function argument provided by the user: `value`*, `field`, `apply`; or from advanced use: `entry`

    Exceptionally, 'field' is a subclass that can also be piped through '>>'
    *value also inherits AbsBaseArgument
    """


class AbsMetaArgument(AbsArgument):
    """
    Special temporary argument-like directive provided by the user: `default`, `sample`
    """



from __future__ import annotations

from itertools import chain

from hdict.data.frozenhdict import frozenhdict
from hosh import ø

from hdict.content.argument import AbsArgument
from hdict.content.argument.apply import apply
from hdict.content.argument.default import default
from hdict.content.entry import AbsEntry, Unevaluated
from hdict.content.entry.aux_closure import handle_arg
from hdict.text.customjson import truncate


class Closure(AbsEntry):
    def __init__(self, application: apply, data: dict[str, AbsEntry], out: list, previous: frozenhdict):
        from hdict.data.aux_frozendict import handle_item

        self.application = application
        self.out = out
        self.torepr = {}
        hosh = ø
        arg = None
        fargs, fkwargs, discarded_defaults = application.fargs.copy(), {}, set()  # We copy fargs to keep args order.
        sortable_fargs = zip(map(str, fargs), fargs.items())
        for idx, tup in sorted(chain(sortable_fargs, application.fkwargs.items())):  # We sort by keys for a deterministic hosh.
            if isinstance(tup, tuple):
                key, val = tup
                arg = handle_arg(key, val, data, discarded_defaults, out, self.torepr, previous)
                fargs[key] = arg
            else:
                key, val = idx, tup
                arg = handle_arg(key, val, data, discarded_defaults, out, self.torepr, previous)
                fkwargs[key] = arg
            hosh *= arg.hosh

        if application.isfield:
            appliable_entry = handle_item(application.appliable.name, application.appliable, data, previous)
            hosh *= appliable_entry.hosh.rev

            def f():
                args = (x.value for x in fargs.values())
                kwargs = {k: v.value for k, v in fkwargs.items()}
                try:
                    return appliable_entry.value(*args, **kwargs)
                except TypeError as e:  # pragma: no cover
                    if "required positional argument" in str(e):
                        raise Exception(f"{str(e)}\n" f"HINT: If you are applying a field, " f"you should make the function parameters explicit, e.g.: `apply(f, parameter1, parameter2)`.")

        else:
            hosh *= application.ahosh
            appliable_function = application.appliable

            def f():
                args = (x.value for x in fargs.values())
                kwargs = {k: v.value for k, v in fkwargs.items()}
                return appliable_function(*args, **kwargs)

        self.f = f
        self.hosh = hosh
        self.discarded_defaults = discarded_defaults

    @property
    def value(self):
        if isinstance(self._value, Unevaluated):
            self._value = self.f()
            # del self.application  # todo: : delete clasure.application inside each subvalue?
        return self._value

    def __repr__(self, out=None):
        from hdict import value
        from hdict import field
        from hdict.content.entry.wrapper import Wrapper

        if out is None:
            out = []
        out = out + self.out
        lst = []
        for param, content in self.torepr.items():
            pre = "" if isinstance(param, int) else f"{param}="

            match content:
                case value():
                    lst.append(truncate(repr(content), width=7))

                case default() if param in self.discarded_defaults:
                    lst.append(f"{param}")
                case default(value=v):
                    lst.append(f"{param}={v}")

                # case field(name) if name in out:
                #     lst.append(repr(content))
                case field(name) if name == param:
                    lst.append(f"{param}")
                case field(name):
                    lst.append(f"{pre}{name}")

                case Wrapper(entry_):
                    if isinstance(entry_, Closure):
                        lst.append(entry_.__repr__(out))
                    else:
                        lst.append(param)
                case Closure():
                    lst.append(f"{pre}{content.value if content.isevaluated else content.__repr__(out)}")
                case AbsArgument():
                    lst.append(f"{pre}{repr(content)}")
                case AbsEntry():
                    lst.append(f"{pre}{repr(content)}")
                case _:  # pragma: no cover
                    raise Exception(type(content))
        return f"λ({' '.join(lst)})"



from indexed import IndexedOrderedDict

from hdict.content.argument import AbsArgument


class IndexedDict(IndexedOrderedDict):
    """
    Wrapper with a cleaner repr(esentation) of IndexedOrderedDict.

    >>> IndexedDict([('a', 'a'), ('b', 'b')])
    {'a': 'a', 'b': 'b'}
    """

    def __repr__(self):
        return repr(dict(self.items()))


class Arg(str):
    """
    >>> Arg("a") >= Arg("a")
    True
    """

    def __init__(self, position: int):
        self.position = position

    def __hash__(self):
        return hash(self.position)

    def __lt__(self, other):
        return f"~{self.position}" < str(other)

    def __ge__(self, other):
        return f"~{self.position}" >= str(other)

    def __repr__(self):
        return f"arg_{self.position}"


def handle_args(signature, applied_args, applied_kwargs):
    from hdict.content.argument.field import field
    from hdict.content.value import value

    # Separate positional from named parameters of 'f'.
    fargs, fkwargs = IndexedDict(), {}
    params = []
    hasargs, haskwargs = False, False
    if signature is None:
        # Applying a field.
        for i, v in enumerate(applied_args):
            fargs[i] = v
            params.append(i)
        for k, v in applied_kwargs.items():
            fkwargs[k] = v
            params.append(k)
    else:
        for par in signature.parameters.values():
            strpar = str(par)
            if strpar.startswith("**"):
                haskwargs = True
                continue
            elif strpar.startswith("*"):
                hasargs = True
                continue
            name = par.name
            if (v := par.default) is par.empty:
                fargs[name] = field(name)
            else:
                from hdict.content.argument.default import default

                fkwargs[name] = default(v)
            params.append(name)

    # apply's entry override f's entry
    wrap = lambda x: x if isinstance(x, AbsArgument) else value(x)
    used = set()
    for i, applied_arg in enumerate(applied_args):
        if i < len(fargs):
            # noinspection PyUnresolvedReferences
            used.add(key := fargs.keys()[i])
            fargs[key] = wrap(applied_arg)
        else:
            if i >= len(params):
                if not hasargs:  # pragma: no cover
                    raise Exception("Too many arguments to apply. No '*arg' detected for 'f'.")
                name = Arg(i)  # REMINDER: this is just a unique hash for the nameless argument
            else:
                name = Arg(i) if hasargs and i >= len(fargs) else params[i]
                if name in fkwargs:
                    del fkwargs[name]
            fargs[name] = wrap(applied_arg)

    for applied_kwarg, v in applied_kwargs.items():
        if applied_kwarg in used:  # pragma: no cover
            raise Exception(f"Parameter '{applied_kwarg}' cannot appear in both 'arg' and 'kwargs' of 'apply()'.")
        if applied_kwarg in fargs:
            fargs[applied_kwarg] = wrap(v)
        elif applied_kwarg in fkwargs or haskwargs:
            fkwargs[applied_kwarg] = wrap(v)
        else:  # pragma: no cover
            raise Exception(f"Parameter '{applied_kwarg}' is not present in 'f' signature nor '**kwargs' was detected for 'f'.")

    return fargs, fkwargs



#  Copyright (c) 2023. Davi Pereira dos Santos
#  This file is part of the hdict project.
#  Please respect the license - more about this in the section (*) below.
#
#  hdict is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  hdict is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with hdict.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is illegal and it is unethical regarding the effort and
#  time spent here.
#

from random import Random

from lange.tricks import list2progression

from hdict.content.value import value
from ..argument import AbsMetaArgument


class sample(AbsMetaArgument):
    """
    A lazily evaluated list of values

    >>> (s := sample(1, 2, 3, ..., 9).values)
    [1 2 .+. 9]
    >>> (s := sample(2, 4, 8, ..., 1024).values)
    [2 4 .*. 1024]
    >>> (s := sample(2, -4, 8, ..., 12).values)
    [2 -4 8]
    """

    sampleable = True

    def __init__(self, *values: list[int | float], rnd: int | Random = 0, maxdigits=28):
        self.rnd = rnd
        # todo: : accept list of non numeric types (categoric)?
        prog = list2progression(values, maxdigits=maxdigits)
        if prog.n.is_infinite():  # pragma: no cover
            raise Exception(f"Cannot sample from an infinite list: {prog}")
        self.values = prog

    def sample(self, rnd: int | Random = None):
        if rnd is None:
            rnd = self.rnd
        if isinstance(rnd, int):
            rnd = Random(rnd)
        if not isinstance(rnd, Random):  # pragma: no cover
            raise Exception(f"Sampling needs an integer seed or a Random object.")
        idx = rnd.randint(0, self.values.n - 1)
        return value(self.values[idx])

    def __repr__(self):
        return f"~{self.values}"



class AbsAny:
    """
    Any class defined in this project that is made to be a hdict value or should not be used as such

    This intended to ease checking disallowed types being mistakengly used as user values.
    """



#  Copyright (c) 2023. Davi Pereira dos Santos
#  This file is part of the hdict project.
#  Please respect the license - more about this in the section (*) below.
#
#  hdict is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  hdict is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with hdict.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is illegal and it is unethical regarding the effort and
#  time spent here.
#

import dis
import re
from collections import OrderedDict
from inspect import signature, isbuiltin, isclass
from pickle import dumps

from hosh import Hosh, ø


def v2hosh(value: object) -> Hosh:
    """
    Convert any value to a hosh object.

    `value` should be serializable (by pickle).
    Adopt pickle for hoshfication because it is faster.

    >>> obj = {"x": 3, "l": [1, 2, "5"]}
    >>> print(v2hosh(obj))
    z0cJsCBPY7cHt.oJnrpyk23FOlPdCcYvcX8x7jg6
    >>> # We encapsulate 'obj' as built-in types cannot be easily patched.
    >>> obj = OrderedDict(obj)
    >>> obj.hosh = ø * "My-custom-identifier-arbitrarily-defined"
    >>> print(v2hosh(obj))
    My-custom-identifier-arbitrarily-defined
    """
    if hasattr(value, "hosh"):
        return value.hosh
    else:
        try:
            if callable(value):
                return f2hosh(value)
            # REMINDER: pickle is the fastest serialization
            return Hosh(dumps(value, protocol=5))
        except TypeError as e:  # pragma: no cover
            raise Exception(f"Cannot pickle. Pickling is needed to hosh hdict values ({value}): {e}")


def f2hosh(function: callable):
    """
    Convert a function to a hosh object.

    Adopt pickle(bytecode) for hoshfication because it is faster than other serializations of bytecode.

    >>> fun = lambda x, y: x + y
    >>> print(f2hosh(fun))
    vD8.I-NU3x5hzmj-m1EJgeAIE-.H.HGnWxqvZng0
    >>> fun.hosh = ø * "My-custom-identifier-arbitrarily-defined"
    >>> print(f2hosh(fun))
    My-custom-identifier-arbitrarily-defined
    """
    if hasattr(function, "hosh"):
        return function.hosh
    if isbuiltin(function) or isclass(function):
        return Hosh(str(function).encode())
    fields_and_params = signature(function).parameters.values()
    fields_and_params = {v.name: None if v.default is v.empty else v.default for v in fields_and_params}

    # Remove line numbers.
    groups = [l for l in dis.Bytecode(function).dis().split("\n\n") if l]
    clean_lines = [fields_and_params]

    for group in groups:
        # Replace memory addresses and file names by just the object name.
        group = re.sub(r'<code object (.+) at 0x[0-f]+, file ".+", line \d+>', r"\1", group)
        lines = [re.sub(r"^[\d ]+", "", segment) for segment in re.split(" +", group)][1:]
        clean_lines.append(lines)
    code = [fields_and_params, clean_lines]
    # REMINDER: pickle chosen because it is the fastest serialization (see bottom of the file)
    return Hosh(dumps(code, protocol=5))


# Timing:
"""
l = [{"x": 3}, 530934590.435903475, "4p9fj24gifh2430g8h230g82h34g0p2843hg02h"] * 100


def f():
    pack(l, ensure_determinism=True, unsafe_fallback=False, compressed=False)


def g():
    dumps(l)


def h():
    pack(l, ensure_determinism=True, unsafe_fallback=False, compressed=True)


print(timeit(f, number=1000))
print(timeit(g, number=1000))
print(timeit(h, number=1000))
"""



from dataclasses import dataclass

from hdict.content.entry import AbsEntry, Unevaluated


@dataclass
class Wrapper(AbsEntry):
    """
    The only entry that can nest and return another entry as a value

    For advanced usage only
    """

    entry: AbsEntry

    def __post_init__(self):
        self.hosh = self.entry.hosh

    @property
    def value(self):
        if isinstance(self._value, Unevaluated):
            self._value = self.entry
        return self._value

    def __repr__(self):
        return "·" + repr(self.entry)



#  Copyright (c) 2023. Davi Pereira dos Santos
#  This file is part of the hdict project.
#  Please respect the license - more about this in the section (*) below.
#
#  hdict is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  hdict is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with hdict.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is illegal and it is unethical regarding the effort and
#  time spent here.
#
from typing import TypeVar

from hdict.content.argument.apply import apply
from hdict.content.argument.field import field
from hdict.content.argument.sample import sample
from hdict.data.empty_ import Empty_
from hdict.content.value import value
from hdict.data.frozenhdict import frozenhdict
from hdict.data.hdict_ import hdict_
from hdict.data.self_ import Self_
from hdict.expression.step.cache import cache

__all__ = ["hdict", "_", "Ø", "apply", "field", "sample", "frozenhdict", "value", "cache"]

VT = TypeVar("VT")


class hdict(hdict_):
    """
    Function id is reversed before application.
    This is needed to enable handling a function as a value, under the original id.

    >>> from hdict import hdict, apply, field
    >>> from hdict.content.argument.default import default
    >>> d = hdict({"x": 3}, y=5)
    >>> d["alpha"] = 11
    >>> d.show(colored=False)
    {
        x: 3,
        y: 5,
        alpha: 11,
        _id: e6WGyTj7trk6AUUcdxGAefyG.T9j.SvBdEsHF.r5,
        _ids: {
            x: KGWjj0iyLAn1RG6RTGtsGE3omZraJM6xO.kvG5pr,
            y: ecvgo-CBPi7wRWIxNzuo1HgHQCbdvR058xi6zmr2,
            alpha: A1RyW.GoA3q9-iCbCvWyWClExm1J3wI.ok6UA3Nk
        }
    }
    >>> d >>= {"beta": 12, "gamma": 17}
    >>> d["p1", "q1"] = apply(lambda x, y=3: [x**y, x/y])
    >>> d["pq"] = apply(lambda x, y=3: [x**y, x/y])
    >>> d["p2", "q2"] = apply(lambda x=2, y=None: [x**y, x/y])
    >>> f = lambda x, y: {"a": x**y, "b": x/y}
    >>> d["z1", "w1"] = apply(f)
    >>> d["z2":"a", "w2":"b"] = apply(f)
    >>> d >>= {
    ...     "r1": apply(lambda x: x**2),
    ...     "r2": apply(lambda x: x**3),
    ...     "r3": apply(lambda x: x**4),
    ...     ("ra1", "rb1"): apply(f),
    ...     (("ra2", "a"), ("rb2", "b")): apply(f),
    ... }
    >>> d["z2"]
    243
    >>> d["zz1", "ww1"] = apply(f, field("r1"), y=field("r2"))
    >>> d >>= {("zz2", "ww2"): apply(f, y=field("r2"), x=9)}  # Define external value.
    >>> d >>= {"zzzz": apply(f, y=13, x=9)}
    >>> # Non-pickable custom classes need a custom 'hosh' attribute to be applied and also to be used as a value.
    >>> from hosh import Hosh
    >>> # Example of class that could be used as a value or as a function.
    >>> from dataclasses import dataclass
    >>> @dataclass
    ... class CustomClass:
    ...     hosh = Hosh.fromid("Some.arbitrary.identifier.with.length.40")
    ...     def __call__(self, x, y):
    ...         return x + y, x / y
    >>> custom = CustomClass()
    >>> d["f"] = custom  # Custom callable handled as a value.
    >>> d["zzz1", "www1"] = apply(custom, 11, y=33)  # Custom callable being applied.
    >>> d["zzz2", "www2"] = apply(field("f"), field("r1"), y=44)  # Callable field being applied.
    >>> d.show(colored=False)
    {
        x: 3,
        y: 5,
        alpha: 11,
        beta: 12,
        gamma: 17,
        p1: λ(x y)→0,
        q1: λ(x y)→1,
        pq: λ(x y),
        p2: λ(x y)→0,
        q2: λ(x y)→1,
        z1: λ(x y)→0,
        w1: λ(x y)→1,
        z2: 243,
        w2: λ(x y)→b,
        r1: λ(x),
        r2: λ(x),
        r3: λ(x),
        ra1: λ(x y)→0,
        rb1: λ(x y)→1,
        ra2: λ(x y)→a,
        rb2: λ(x y)→b,
        zz1: λ(x=r1 y=r2)→0,
        ww1: λ(x=r1 y=r2)→1,
        zz2: λ(9 y=r2)→0,
        ww2: λ(9 y=r2)→1,
        zzzz: λ(9 13),
        f: "CustomClass()",
        zzz1: λ(11 33)→0,
        www1: λ(11 33)→1,
        zzz2: λ(r1 44)→0,
        www2: λ(r1 44)→1,
        _id: i5wtTs5qggivS-y5TNuhnM5jwokjF132M7jF.49a,
        _ids: {
            x: KGWjj0iyLAn1RG6RTGtsGE3omZraJM6xO.kvG5pr,
            y: ecvgo-CBPi7wRWIxNzuo1HgHQCbdvR058xi6zmr2,
            alpha: A1RyW.GoA3q9-iCbCvWyWClExm1J3wI.ok6UA3Nk,
            beta: WM5TbiaJ1gLqKRSFiY3VkZEu1PwQcaAokBKBPrWg,
            gamma: yb-HU.jSxd496XfId1J..MkX7xfUPJOL1-07hHdt,
            p1: Qs0J0I.AhR.brQjpYwUJqDNLk-zKd7FmG0g7gdd1,
            q1: VOAVwyrrdC7eIsjnVkqdpsqUfhprKzKriATFEOQI,
            pq: KtTWRcFQRKVgTJabGMgY-HAQWJ1TZGnOYw7NU.1K,
            p2: qb0H0MqNit0rLwk6CsC4Q2bxpCmBQ581eoGI39sr,
            q2: Sjsl-.MJJ3SK2.INLAJ-0R5BwlfG.tccIpxPFRtB,
            z1: w8obIVknqdoKk.hDUN8TE2M-KiFh-yCCXtxNdrJ4,
            w1: T0hCevYdBG6vWZ3R5Nba.jXacMTFFUW5OjOX4yAO,
            z2: w8obIVknqdoKk.hDUN8TE2M-KiFh-yCCXtxNdrJ4,
            w2: T0hCevYdBG6vWZ3R5Nba.jXacMTFFUW5OjOX4yAO,
            r1: FO3NXIQIi7TBFw1FqLRJYo13EA.uu4Y-L17Fmx0z,
            r2: HorxHsNCdpkm270yeTZ-vREjjFXWIYC99ALidDVU,
            r3: bWc33lXvXHIJXix5Jw2.Nhw0EIsv3TuAa7t8Ix5q,
            ra1: w8obIVknqdoKk.hDUN8TE2M-KiFh-yCCXtxNdrJ4,
            rb1: T0hCevYdBG6vWZ3R5Nba.jXacMTFFUW5OjOX4yAO,
            ra2: w8obIVknqdoKk.hDUN8TE2M-KiFh-yCCXtxNdrJ4,
            rb2: T0hCevYdBG6vWZ3R5Nba.jXacMTFFUW5OjOX4yAO,
            zz1: FB6W9j12qWZ34fUwyrBETzFuAXRB-xn7O1PGhr2j,
            ww1: IhchjkRfjLXpnsC6zl-TFugIQrnUIdB7Wsgpnely,
            zz2: LE5ohPby7q4dNu3qhqgucRcCt.pd87O.CyKWloqe,
            ww2: -RBf-UiccAG3sTH0UFC6C.YFmYcpTRA.0PFP6Oun,
            zzzz: Irbj-xVP8rEIPoylOTNmiuJ-lGdx0dpzIPgq-169,
            f: Some.arbitrary.identifier.with.length.40,
            zzz1: TBw0fUZZo1WlI9uczNuF0w4vC06u6YvzJyNPVEi1,
            www1: WqqXpIpMwinaXxYNqRjj4qttDcKXN.gcGjARpOSh,
            zzz2: .mJim0y-LMUYZi01GN8MTPkHj79JtFdREf7g40ug,
            www2: jwtsaf1IZMG6sWqprXZnc.n-ilyfncwhL3Qr.-Md
        }
    }
    >>> d["p1"]
    243
    >>> from hdict import value
    >>> d.evaluate()
    >>> d = hdict(x=2)
    >>> g = lambda x, y: [x + y, x / y]
    >>> d["z"] = apply(f, 2, y=3)
    >>> d["w", "v"] = apply(f, field("x"), y=33)
    >>> d["f"] = f
    >>> d = d >> apply(field("f"), field("x"), y=default(3), nonexistent_parameter=7)("w3", "v3") # todo: : nonexistent parameter should raise an exception
    >>> d >>= apply(field("f"), field("x"), y=default(3))("w", "v")
    >>> d["w2", "v2"] = apply(field("f"), field("x"), y=default(3))
    >>> d >>= {"z": apply(f, field("x"), y=3), ("w", "v"): apply(g, y=7)}
    >>> d >>= apply(f, field("x"), y=3)("z9") * apply(g, y=7)("w9", "v9")
    >>> pp = apply(f, field("x"), y=3)("z") >> apply(g, y=7)("w", "v")
    >>> d >>= {"x": 3} >> pp >> apply(g, y=7)("w", "v")
    >>> from hdict import _
    >>> a1 = apply(f, y=_(1, 2, 4, ..., 128))
    >>> a2 = apply(f, _(0, 3, 6, ..., 9), y=_(0, 3, 6, ..., 9))
    >>> ppp = hdict() >> a2.sample()("k", "t")
    >>> ppp.show(colored=False)
    {
        k: λ(9 9)→0,
        t: λ(9 9)→1,
        _id: dK-iumb4L5nkFVT9LCkdk3daQtuC.ORk6JwWMhnt,
        _ids: {
            k: rHTgwW.w2SsbcvBlSnWWddLH4vgyXlM1Fhxqr48f,
            t: IldA8m-uSN9lJcE4TBtjTcY-sSiA33mzxQ2k-wpN
        }
    }
    >>> ppp.k
    387420489
    >>> a1("z")
    z=λ(x y=~[1 2 .*. 128])
    >>> a2(w="a", v="b")
    (('w', 'a'), ('v', 'b'))=λ(x=~[0 3 .+. 9] y=~[0 3 .+. 9])
    >>> p = a1("z") >> a2(w="a", v="b")
    >>> h = lambda a, b=4: 5
    >>> app = apply(h, a=_(0, 3, 6, ..., 9))
    >>> app
    λ(a=~[0 3 .+. 9] b=default(4))
    >>> app.c
    c=λ(a=~[0 3 .+. 9] b=default(4))
    >>> sampled = app.sample(0).c
    >>> sampled
    c=λ(9 b=default(4))
    >>> r = hdict() >> sampled
    >>> r.show(colored=False)
    {
        c: λ(9 b=4),
        _id: Bbe5mzn0oCuHZ.Y84-jaEnh3jcrgD8av6IZezKEG,
        _ids: {
            c: 6EehD7OoZe2REI.1YsVLKxSF4hJoYxfOZJLbF.Aj
        }
    }
    >>> r.evaluated.show(colored=False)
    {
        c: 5,
        _id: Bbe5mzn0oCuHZ.Y84-jaEnh3jcrgD8av6IZezKEG,
        _ids: {
            c: 6EehD7OoZe2REI.1YsVLKxSF4hJoYxfOZJLbF.Aj
        }
    }
    >>> from random import Random
    >>> rnd = Random(0)
    >>> p1 = p.sample(rnd)
    >>> d["w"]
    10
    >>> d >>= p1
    >>> d["z"] = apply(f, 2, y=3)
    >>> d["w", "v"] = apply(f, _.x, y=_.x)
    >>> d["w", "v"] = apply(_.f, _.x, y=default(3))
    >>> d = hdict() >> {"z": apply(f, 7, y=3), ("w", "v"): apply(g, default(6), y=7)}
    >>> d = hdict(w=6) >> (apply(f, _["w"], y=3)(z="z") >> apply(g, x=_(1,2,3,...,5), y=7)("ww", "v")).sample(0)
    >>> p = apply(f, y=_(1, 2, 4, ..., 128))("z") >> apply(f, y=_(0, 3, 6, ..., 9))(w="a", v="b")
    >>> d.show(colored=False)
    {
        w: 6,
        z: λ(x=w 3)→z,
        ww: λ(4 7)→0,
        v: λ(4 7)→1,
        _id: x4TxaVuAymsh97Yr.15Oc1ekllvlpECk091124RM,
        _ids: {
            w: CZ7Jm5fQMZ3fZJ3kAVOi0FYK-exFqPqgoYLsGxGl,
            z: No1nN40OXnJ.zOyAvoxwCjyD06cUDiXTZQh4q8xa,
            ww: qZepp.wpXCOYQrwnVIwoZ4kR9pIuMGDQpvRevv80,
            v: wxkjBSP54cKEfxKE1Zte-2dDll3G9Nd-yDh3CLas
        }
    }
    >>> d = hdict(x=3) >> p.sample(rnd)
    >>> d.show(colored=False)
    {
        x: 3,
        z: λ(x 16),
        w: λ(x 9)→a,
        v: λ(x 9)→b,
        _id: 22b-e.CtRGVSoMkektNzHYSVZTouhL2jAHLyPd4Q,
        _ids: {
            x: KGWjj0iyLAn1RG6RTGtsGE3omZraJM6xO.kvG5pr,
            z: 64cxw0HmJ0fydH7By6i0HrbuwA759XexTu3Bu0Zd,
            w: p1Z5umNo12DHCZtKbh2EClW6TZxv.M447HlALWhv,
            v: HbYbR1RjPdU4jixjW7xSxslRpYx9W6Oiay7maK6F
        }
    }
    >>> d1 = hdict(x=52, y=13)
    >>> d2 = hdict(x=value(52, hosh="1234567890123456789012345678901234567890"))
    >>> d1.show(colored=False)
    {
        x: 52,
        y: 13,
        _id: 5pwGP0LxQRbQaDEk9ztQ4V4qz.A-IIhVNTrcyt8U,
        _ids: {
            x: c.llz05T6ZXw5AlsU4xfmMxcvvHZhLl60LckZis9,
            y: zplslThZYha4haD2VmGxZqrFSw5RJcFJd2E-Ku6s
        }
    }
    >>> d2.show(colored=False)
    {
        x: 52,
        _id: kYzgpPdRgQSYSEpp1qt4EHQLQJXuyb2WDQS-iNPh,
        _ids: {
            x: 1234567890123456789012345678901234567890
        }
    }
    >>> d3 = d1 >> d2
    >>> d3.show(colored=False)
    {
        x: 52,
        y: 13,
        _id: -EFuy5NAeK.LIALpBiZKK-fYQmc9AZYQQck-HiRK,
        _ids: {
            x: 1234567890123456789012345678901234567890,
            y: zplslThZYha4haD2VmGxZqrFSw5RJcFJd2E-Ku6s
        }
    }
    >>> (d3 >> d1.frozen).show(colored=False)
    {
        x: 52,
        y: 13,
        _id: 5pwGP0LxQRbQaDEk9ztQ4V4qz.A-IIhVNTrcyt8U,
        _ids: {
            x: c.llz05T6ZXw5AlsU4xfmMxcvvHZhLl60LckZis9,
            y: zplslThZYha4haD2VmGxZqrFSw5RJcFJd2E-Ku6s
        }
    }

    >>> d = hdict(x=2, y=4)
    >>> {"x": 2, "y": 4} == d.frozen
    True
    >>> {"x": 2, "y": 4} == d
    True
    >>> d == {"x": 2, "y": 4}
    True
    >>> dict(d)
    {'x': 2, 'y': 4}
    >>> hdict() >> {"x": 3} == {"x": 3}
    True
    >>> {"x": 3, "d": {"x": 7}} == hdict(x=3, d=hdict(x=7))
    True
    >>> hdict(x=3, d=hdict(x=7)) == {"x": 3, "d": {"x": 7}}
    True
    >>> {"x": 3, "_id": hdict(x=3).id} == hdict(x=3)
    True
    >>> hdict(x=3) == {"x": 3, "_id": hdict(x=3).id}
    True
    >>> hdict(x=3) == hdict(x=3)
    True
    >>> hdict(x=3).frozen == hdict(x=3)
    True
    >>> hdict(x=3) != {"x": 4}
    True
    >>> hdict(x=3) != hdict(x=4)
    True
    >>> hdict(x=3).frozen != hdict(x=4)
    True
    >>> hdict(x=3) != {"y": 3}
    True
    >>> hdict(x=3) != {"x": 3, "_id": (~hdict(x=3).hosh).id}
    True
    >>> hdict(x=3) != hdict(y=3)
    True
    >>> del d["x"]
    >>> list(d)
    ['y']
    >>> e = d >> apply(lambda y: y*7)("y")
    >>> from hdict.content.aux_value import f2hosh
    >>> print(f2hosh(lambda y: y*7))
    54YCMDJIlsIvMQ.KJtT-vFyjg83Zgfj2xSHOgCj8
    >>> print(e)
    {y: "λ(4)"}
    >>> e.evaluate()
    >>> print(e)
    {y: 28}
    >>> d = d >> apply(lambda y=1: y*7, fhosh="54YCMDJIlsIvMQ.KJtT-vFyjg83Zgfj2xSHOgCj8")("y")
    >>> print(d)
    {y: "λ(4)"}
    >>> d.evaluate()
    >>> print(d)
    {y: 28}
    >>> hash(e.frozen) == hash(d.frozen)
    True
    >>> d = hdict(a=5) >> hdict(y=28)
    >>> d.show(colored=False)
    {
        a: 5,
        y: 28,
        _id: C-xKyWCyBL6g32KIuxoANoF9czLaJTh-emPsMqOg,
        _ids: {
            a: ecvgo-CBPi7wRWIxNzuo1HgHQCbdvR058xi6zmr2,
            y: -2A0hTRBN1wtIKQxLzRcYDBkhv1hu-dMY-24Jye9
        }
    }
    >>> d >>= apply(lambda a: a)("x")
    >>> d.show(colored=False)
    {
        a: 5,
        y: 28,
        x: λ(a),
        _id: qI8rmiUzMTSO1pMkCeQJpHTAOw4xuooFqM41iIiY,
        _ids: {
            a: ecvgo-CBPi7wRWIxNzuo1HgHQCbdvR058xi6zmr2,
            y: -2A0hTRBN1wtIKQxLzRcYDBkhv1hu-dMY-24Jye9,
            x: sxkIk6E9l8oScCyoGDlzrqJ9QgpIpl5PCBvHlERp
        }
    }
    >>> {"_id": "qI8rmiUzMTSO1pMkCeQJpHTAOw4xuooFqM41iIiY"} == d
    True
    >>> d.show(colored=False)
    {
        a: 5,
        y: 28,
        x: λ(a),
        _id: qI8rmiUzMTSO1pMkCeQJpHTAOw4xuooFqM41iIiY,
        _ids: {
            a: ecvgo-CBPi7wRWIxNzuo1HgHQCbdvR058xi6zmr2,
            y: -2A0hTRBN1wtIKQxLzRcYDBkhv1hu-dMY-24Jye9,
            x: sxkIk6E9l8oScCyoGDlzrqJ9QgpIpl5PCBvHlERp
        }
    }
    >>> def f():
    ...     print("busy")
    ...     return 23
    >>> storage = {}
    >>> d >>= apply(f).o >> cache(storage, "x", "y")
    >>> d.y
    28
    >>> d.show(colored=False)
    {
        a: 5,
        y: 28,
        x: ↑↓ cached at `dict`·,
        o: λ(),
        _id: 4FdTqXVIMKldUnZgrrp315c78KHD9yu7T.Nr5zGv,
        _ids: {
            a: ecvgo-CBPi7wRWIxNzuo1HgHQCbdvR058xi6zmr2,
            y: -2A0hTRBN1wtIKQxLzRcYDBkhv1hu-dMY-24Jye9,
            x: sxkIk6E9l8oScCyoGDlzrqJ9QgpIpl5PCBvHlERp,
            o: Re1o0YXNebYNezPrVv2dOrFxtgLQutD-b-SHFRbo
        }
    }
    >>> d.evaluated.show(colored=False)
    busy
    {
        a: 5,
        y: 28,
        x: 5,
        o: 23,
        _id: 4FdTqXVIMKldUnZgrrp315c78KHD9yu7T.Nr5zGv,
        _ids: {
            a: ecvgo-CBPi7wRWIxNzuo1HgHQCbdvR058xi6zmr2,
            y: -2A0hTRBN1wtIKQxLzRcYDBkhv1hu-dMY-24Jye9,
            x: sxkIk6E9l8oScCyoGDlzrqJ9QgpIpl5PCBvHlERp,
            o: Re1o0YXNebYNezPrVv2dOrFxtgLQutD-b-SHFRbo
        }
    }
    >>> d.evaluated.show(colored=False)
    {
        a: 5,
        y: 28,
        x: 5,
        o: 23,
        _id: 4FdTqXVIMKldUnZgrrp315c78KHD9yu7T.Nr5zGv,
        _ids: {
            a: ecvgo-CBPi7wRWIxNzuo1HgHQCbdvR058xi6zmr2,
            y: -2A0hTRBN1wtIKQxLzRcYDBkhv1hu-dMY-24Jye9,
            x: sxkIk6E9l8oScCyoGDlzrqJ9QgpIpl5PCBvHlERp,
            o: Re1o0YXNebYNezPrVv2dOrFxtgLQutD-b-SHFRbo
        }
    }
    """


class Empty(Empty_):
    """
    >>> from hdict import _
    >>> d = _ >> {"x": 5} >> dict(y=7)
    >>> type(+_), type(d)
    (<class 'hdict.Empty'>, <class 'hdict.hdict'>)
    >>> d.show(colored=False)
    {
        x: 5,
        y: 7,
        _id: A0G3Y7KNMLihDvpSJ3tB.zxshc6u1CbbiiYjCAAA,
        _ids: {
            x: ecvgo-CBPi7wRWIxNzuo1HgHQCbdvR058xi6zmr2,
            y: eJCW9jGsdZTD6-AD9opKwjPIOWZ4R.T0CG2kdyzf
        }
    }
    """


class Self(Self_):
    """"""


Ø = empty = Empty()
_ = Self()



from dataclasses import dataclass

from hdict.expression.step.step import AbsStep
from hdict.text.customjson import stringfy


@dataclass
class EDict(AbsStep):
    """
    Wrapper for dict inside an Expr object

    Useful to keep the step as a `dict` as long as possible, specially when printing the expression.
    """

    dct: dict

    def __repr__(self):
        return stringfy(self.dct)

    def __bool__(self):
        return bool(self.dct)



#  Copyright (c) 2023. Davi Pereira dos Santos
#  This file is part of the hdict project.
#  Please respect the license - more about this in the section (*) below.
#
#  hdict is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  hdict is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with hdict.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is illegal and it is unethical regarding the effort and
#  time spent here.
#
from dataclasses import dataclass

from hdict.content.argument import AbsBaseArgument


@dataclass
class field(AbsBaseArgument):
    """
    Pointer to a field, without knowing the concrete value yet

    >>> from hdict import Ø, _, apply
    >>> d = Ø
    >>> d.show(colored=False)
    {
        _id: 0000000000000000000000000000000000000000,
        _ids: {}
    }
    >>> d = +_
    >>> d.show(colored=False)
    {
        _id: 0000000000000000000000000000000000000000,
        _ids: {}
    }
    >>> d >>= {"x": 3, "y": 5} >> apply(lambda x, y: x + y).z
    >>> d.show(colored=False)
    {
        x: 3,
        y: 5,
        z: λ(x y),
        _id: ..SSoOQF.FYZDUevP1CGj0DJVc.M8cgimN70yI3C,
        _ids: {
            x: KGWjj0iyLAn1RG6RTGtsGE3omZraJM6xO.kvG5pr,
            y: ecvgo-CBPi7wRWIxNzuo1HgHQCbdvR058xi6zmr2,
            z: 3j029VhuvENAnBK6JLRGl8ePpsQybm57sXKfX2oo
        }
    }
    >>> d >>= {"x": 3, "y": 5} >> apply(lambda x, y: x + y).x
    >>> d.x
    8
    >>> d["x"] = apply(lambda x, y: x + y)
    >>> d.x
    13
    """

    name: str

    def __repr__(self):
        return f"field({self.name})"



#  Copyright (c) 2023. Davi Pereira dos Santos
#  This file is part of the hdict project.
#  Please respect the license - more about this in the section (*) below.
#
#  hdict is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  hdict is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with hdict.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is illegal and it is unethical regarding the effort and
#  time spent here.
#
from dataclasses import dataclass

from hdict.content.entry import AbsEntry


@dataclass
class SubValue(AbsEntry):
    """
    A field containing part of other field

    >>> from hdict import value
    >>> v = SubValue(value([3]), 0, 1)
    >>> v
    [3]→0
    >>> v.value
    3
    >>> v
    3
    """

    parent: AbsEntry
    index: int
    n: int
    source: str = None

    # target: str = None

    def __post_init__(self):
        self.hosh = self.parent.hosh[self.index, self.n]

    @property
    def value(self):
        from hdict.content.entry import Unevaluated

        if isinstance(self._value, Unevaluated):
            value = self.parent.value
            if isinstance(value, (list, tuple)):
                if len(value) < self.n:  # pragma: no cover
                    raise Exception(f"Number of output fields ('{self.n}') should not exceed number of resulting list elements ('{len(value)}').")
                self._value = value[self.index]
            elif isinstance(value, dict):
                if len(value) != self.n:  # pragma: no cover
                    raise Exception(f"Number of output fields ('{self.n}') should match number of resulting dict entries ('{len(value)}').")
                if self.source:
                    self._value = value[self.source]
                self._value = list(sorted(value.items()))[self.index][1]
            else:  # pragma: no cover
                raise Exception(f"Cannot infer subvalue '{self.index}' of type '{type(value).__name__} {value}.")
        return self._value

    def __repr__(self):
        if self.isevaluated:
            return repr(self._value)
        # t = self.parent if self.target is None else self.target
        return f"{self.parent}→{str(self.index if self.source is None else self.source)}"



#  Copyright (c) 2021. Davi Pereira dos Santos
#  This file is part of the hdict project.
#  Please respect the license - more about this in the section (*) below.
#
#  hdict is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  hdict is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with hdict.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is illegal and it is unethical regarding the effort and
#  time spent here.
#

# library arff2pandas simply disappeared...
import re


def isplit(source, sep=None, regex=False):  # pragma: no cover
    """
    https://stackoverflow.com/a/9773142/9681577

    generator version of str.split()

    :param source:
        source string (unicode or bytes)

    :param sep:
        separator to split on.

    :param regex:
        if True, will treat sep as regular expression.

    :returns:
        generator yielding elements of string.
    """
    if sep is None:  # pragma: no cover
        # mimic default python behavior
        source = source.strip()
        sep = "\\s+"
        if isinstance(source, bytes):
            sep = sep.encode("ascii")
        regex = True
    if regex:
        # version using re.finditer()
        if not hasattr(sep, "finditer"):
            sep = re.compile(sep)
        start = 0
        for m in sep.finditer(source):
            idx = m.start()
            assert idx >= start
            yield source[start:idx]
            start = m.end()
        yield source[start:]
    else:
        # version using str.find(), less overhead than re.finditer()
        sepsize = len(sep)
        start = 0
        while True:
            idx = source.find(sep, start)
            if idx == -1:
                yield source[start:]
                return
            yield source[start:idx]
            start = idx + sepsize


def liac2pandas(arff):
    import pandas as pd

    attrs = arff["attributes"]
    attrs_t = []
    for attr in attrs:
        if isinstance(attr[1], list):
            attrs_t.append("%s@{%s}" % (attr[0], ",".join(attr[1])))
        else:
            attrs_t.append("%s@%s" % (attr[0], attr[1]))

    df = pd.DataFrame(data=arff["data"], columns=attrs_t)
    return df


def load(fp):
    import arff as liacarff

    data = liacarff.load(fp)
    return liac2pandas(data)


def loads(s):
    import arff as liacarff

    data = liacarff.loads(s)
    return liac2pandas(data)


def df2liac(df, relation="data", description=""):  # pragma: no cover
    attrs = []
    for col in df.columns:
        attr = col.split("@")
        if attr[1].count("{") > 0 and attr[1].count("}") > 0:
            vals = attr[1].replace("{", "").replace("}", "").split(",")
            attrs.append((attr[0], vals))
        else:
            attrs.append((attr[0], attr[1]))

    data = list(df.values)
    result = {"attributes": attrs, "data": data, "description": description, "relation": relation}
    return result


# def dump(df, fp):  # todo: : save arff/CSV from hdict
#     import arff as liacarff
#
#     arff = df2liac(df)
#     liacarff.dump(arff, fp)
#
#
# def dumps(df):  # todo: : output arff/CSV from hdict
#     import arff as liacarff
#
#     arff = df2liac(df)
#     return liacarff.dumps(arff)


def df2Xy(df, target=None):
    r"""
    >>> from hdict import hdict
    >>> from hdict.dataset.dataset import df2Xy
    >>> from testfixtures import TempDirectory
    >>> arff = "@RELATION mini\n@ATTRIBUTE attr1	REAL\n@ATTRIBUTE attr2 	REAL\n@ATTRIBUTE class 	{0,1}\n@DATA\n5.1,3.5,0\n3.1,4.5,1"
    >>> with TempDirectory() as tmp:  # doctest:+ELLIPSIS
    ...    tmp.write("mini.arff", arff.encode())
    ...    d = hdict.fromfile(tmp.path + "/mini.arff")
    '/tmp/.../mini.arff'
    >>> df2Xy(d.df)
    {'X':    attr1  attr2
    0    5.1    3.5
    1    3.1    4.5, 'y': array([0, 1])}
    >>> df2Xy(d.df, target="attr2")
    {'X':    attr1 class
    0    5.1     0
    1    3.1     1, 'y': 0    3.5
    1    4.5
    Name: attr2, dtype: float64}
    """
    if target is None:
        from sklearn.preprocessing import LabelEncoder

        le = LabelEncoder()
        X = df.drop(df.columns[[-1]], axis=1)
        y = le.fit_transform(df[df.columns[-1]])
    else:
        y = df[target]
        X = df.drop(target, axis=1)
    return {"X": X, "y": y}


def nom2bin(X, nomcols):
    """
    >>> import numpy as np
    >>> from pandas import DataFrame as DF
    >>> X = DF(np.array([[0, "a", 1.6], [3.2, "b", 2], [8, "c", 3]]))
    >>> X
         0  1    2
    0    0  a  1.6
    1  3.2  b    2
    2    8  c    3
    >>> nom2bin(X, nomcols=[1])
         0    2  1_a  1_b  1_c
    0    0  1.6    1    0    0
    1  3.2    2    0    1    0
    2    8    3    0    0    1
    """
    if X.__class__.__name__ in ["DataFrame", "Series"]:
        import pandas

        clabels = X.columns[nomcols]
        return pandas.get_dummies(X, prefix=clabels, columns=clabels, dtype=int)
    else:
        import numpy
        from sklearn.preprocessing import OneHotEncoder

        encoder = OneHotEncoder()
        nom = encoder.fit_transform(X.iloc[:, nomcols] if hasattr(X, "iloc") else X[:, nomcols]).toarray()
        num = numpy.delete(X, nomcols, axis=1).astype(float)
        return numpy.column_stack((nom, num))


#
#
# def df2arff(input="df", output="arff", **kwargs):
#     """
#     >>> from idict import let, idict
#     >>> d = idict.fromminiarff()
#     >>> d >>= let(df2arff, output="a")
#     >>> d.show(colored=False)
#     {
#         "a": "→(input output df)",
#         "_history": "idict---------arff2pandas-1.0.1--df2arff",
#         "df": "«{'attr1@REAL': {0: 5.1, 1: 3.1}, 'attr2@REAL': {0: 3.5, 1: 4.5}, 'class@{0,1}': {0: '0', 1: '1'}}»",
#         "_id": "Ojq9k7ZSbVjwLGlZ7uuqIyhoJPo.p36mAmav2Wul",
#         "_ids": {
#             "a": "wVPgESiPwobfFLLm591BKw6i2Zn.p36mAmav2Wul",
#             "_history": "D2NpAYrhyJW-.nOhh91ttSjrbGw.nZ8qxps9giws",
#             "df": "q3_b71eb05c4be05eba7b6ae5a9245d5dd70b81b (content: 6X_dc8ccea3b2e46f1c78967fae98b692701dc99)"
#         }
#     }
#     >>> d.a
#     '@RELATION data\\n\\n@ATTRIBUTE attr1 REAL\\n@ATTRIBUTE attr2 REAL\\n@ATTRIBUTE class {0, 1}\\n\\n@DATA\\n5.1,3.5,0\\n3.1,4.5,1\\n'
#     """
#     return {output: dumps(kwargs[input]), "_history": ...}
#
#
# def openml(Xout="X", yout="y", name="iris", version=1):
#     """
#     #>>> from idict import Ø
#     #>>> (Ø >> openml).show(colored=False)
#     #{
#         #"X": "→(Xout yout name version)",
#         #"y": "→(Xout yout name version)",
#         #"_history": "idict--------------sklearn-1.0.1--openml",
#         #"_id": "idict--------------sklearn-1.0.1--openml",
#         #"_ids": {
#             #"X": "KkSoAvgPmGq52PvPBqGFCEp9cSUCfb7eht9VpUdr",
#             #"y": "HOE9E-HnFYen6JNRC9cHgAamJlOWGYTPIVKvQu8W",
#             #"_history": "Efw0ebrPTxTiuJ.ZpVt-MvB62ja.kmrOYzrn-ljf"
#         #}
#     #}
#     #>>> (Ø >> openml).X.head()
#        #sepallength  sepalwidth  petallength  petalwidth
#     #0          5.1         3.5          1.4         0.2
#     #1          4.9         3.0          1.4         0.2
#     #2          4.7         3.2          1.3         0.2
#     #3          4.6         3.1          1.5         0.2
#     #4          5.0         3.6          1.4         0.2
#     #>>> (Ø >> openml).y.head()
#     #0    Iris-setosa
#     #1    Iris-setosa
#     #2    Iris-setosa
#     #3    Iris-setosa
#     #4    Iris-setosa
#     #Name: class, dtype: category
#     #Categories (3, object): ['Iris-setosa', 'Iris-versicolor', 'Iris-virginica']
#     """
#     from sklearn.datasets import fetch_openml
#
#     X, y = fetch_openml(name=name, version=version, as_frame=True, return_X_y=True)
#     return {Xout: X, yout: y, "_history": ...}
#
#



#  Copyright (c) 2023. Davi Pereira dos Santos
#  This file is part of the hdict project.
#  Please respect the license - more about this in the section (*) below.
#
#  hdict is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  hdict is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with hdict.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is illegal and it is unethical regarding the effort and
#  time spent here.
#
from dataclasses import dataclass

from hdict.content.argument import AbsBaseArgument


@dataclass
class entry(AbsBaseArgument):
    """
    Pointer to an entry, a non evaluated lower level than `field`
    """

    name: str

    def __repr__(self):
        return f"·{self.name}"



#  Copyright (c) 2023. Davi Pereira dos Santos
#  This file is part of the hdict project.
#  Please respect the license - more about this in the section (*) below.
#
#  hdict is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  hdict is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with hdict.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is illegal and it is unethical regarding the effort and
#  time spent here.
#

import json
import re
from collections import UserDict
from io import StringIO
from typing import TypeVar, Union

from hdict.dataset.dataset import loads, isplit
from hdict.dataset.pandas_handling import file2df
from hdict.text.customjson import CustomJSONEncoder, stringfy

VT = TypeVar("VT")


class frozenhdict(UserDict, dict[str, VT]):
    """
    Immutable hdict.

    Any nested 'hdict' value will be frozen to avoid inconsistency between the hdict id (inner id) and the frozenhdict id (outer id).

    >>> from hdict import frozenhdict, hdict
    >>> d = frozenhdict({"x": 3}, y=5)
    >>> from hosh._internals_appearance import decolorize
    >>> print(decolorize(repr(d)))  # This is equivalent to just 'd', without colors.
    {
        x: 3,
        y: 5,
        _id: r5A2Mh6vRRO5rxi5nfXv1myeguGSTmqHuHev38qM,
        _ids: {
            x: KGWjj0iyLAn1RG6RTGtsGE3omZraJM6xO.kvG5pr,
            y: ecvgo-CBPi7wRWIxNzuo1HgHQCbdvR058xi6zmr2
        }
    }
    >>> bool(d), bool(frozenhdict())
    (True, False)
    >>> d.data
    {'x': 3, 'y': 5}
    >>> from hdict import _, apply
    >>> d *= apply(lambda v, x: v - x).z
    >>> str(d)
    '⦑{x: 3, y: 5} » z=λ(v x)⦒'
    >>> d.show(colored=False)
    ⦑{
        x: 3,
        y: 5,
        _id: r5A2Mh6vRRO5rxi5nfXv1myeguGSTmqHuHev38qM,
        _ids: {
            x: KGWjj0iyLAn1RG6RTGtsGE3omZraJM6xO.kvG5pr,
            y: ecvgo-CBPi7wRWIxNzuo1HgHQCbdvR058xi6zmr2
        }
    } » z=λ(v x)⦒
    >>> d = {"v": 7} * d
    >>> d.solve().show(colored=False)
    {
        v: 7,
        x: 3,
        y: 5,
        z: λ(v x),
        _id: 0qFPOnULZigyiXU9Jv.1D.XSTIYHZ24UCT00DMHF,
        _ids: {
            v: eJCW9jGsdZTD6-AD9opKwjPIOWZ4R.T0CG2kdyzf,
            x: KGWjj0iyLAn1RG6RTGtsGE3omZraJM6xO.kvG5pr,
            y: ecvgo-CBPi7wRWIxNzuo1HgHQCbdvR058xi6zmr2,
            z: 2-wfv1b9RyFHB2kdba3EBCy6Do5L-mMPJjT-nfPa
        }
    }
    >>> d = _ >> d
    >>> d.show(colored=False)
    {
        v: 7,
        x: 3,
        y: 5,
        z: λ(v x),
        _id: 0qFPOnULZigyiXU9Jv.1D.XSTIYHZ24UCT00DMHF,
        _ids: {
            v: eJCW9jGsdZTD6-AD9opKwjPIOWZ4R.T0CG2kdyzf,
            x: KGWjj0iyLAn1RG6RTGtsGE3omZraJM6xO.kvG5pr,
            y: ecvgo-CBPi7wRWIxNzuo1HgHQCbdvR058xi6zmr2,
            z: 2-wfv1b9RyFHB2kdba3EBCy6Do5L-mMPJjT-nfPa
        }
    }
    >>> d = {"a": 5} >> d
    >>> d.show(colored=False)
    {
        a: 5,
        v: 7,
        x: 3,
        y: 5,
        z: λ(v x),
        _id: jZrJ2CiVUA9OqW.G7dwb3KoaTWGMUmSUVUXn0CkM,
        _ids: {
            a: ecvgo-CBPi7wRWIxNzuo1HgHQCbdvR058xi6zmr2,
            v: eJCW9jGsdZTD6-AD9opKwjPIOWZ4R.T0CG2kdyzf,
            x: KGWjj0iyLAn1RG6RTGtsGE3omZraJM6xO.kvG5pr,
            y: ecvgo-CBPi7wRWIxNzuo1HgHQCbdvR058xi6zmr2,
            z: 2-wfv1b9RyFHB2kdba3EBCy6Do5L-mMPJjT-nfPa
        }
    }
    >>> from hdict.content.entry import AbsEntry, Unevaluated
    >>> from hdict import frozenhdict
    """

    _evaluated = None
    _asdict, _asdicts, _asdicts_noid = None, None, None
    _hoshes = None

    # noinspection PyMissingConstructor
    def __init__(self, /, _dictionary=None, _previous=None, **kwargs):
        from hdict.content.entry import AbsEntry
        from hdict.data.aux_frozendict import handle_identity
        from hdict.data.aux_frozendict import handle_items

        # todo: : check if _dictionary keys is 'str'; regex to check if k is an identifier;
        data = _dictionary or {}
        # REMINDER: Inside data, the only 'dict' entries are "_id" and "_ids", the rest are AbsEntry objects.
        self.data: dict[str, AbsEntry | str | dict[str, str]]
        self.data = handle_items(data, kwargs, previous=_previous)
        self.hosh, self.ids = handle_identity(self.data)
        self.id = self.hosh.id
        self.raw = self.data

    @property
    def hoshes(self):
        if self._hoshes is None:
            self._hoshes = {k: v.hosh for k, v in self.data.items()}
        return self._hoshes

    def __rmul__(self, left):
        from hdict import frozenhdict
        from hdict.expression.step.edict import EDict
        from hdict.expression.expr import Expr

        from hdict import hdict

        if isinstance(left, dict) and not isinstance(left, (hdict, frozenhdict)):
            return Expr(EDict(left), self)
        return NotImplemented  # pragma: no cover

    def __mul__(self, other):
        from hdict import hdict, frozenhdict
        from hdict.expression.step.edict import EDict
        from hdict.expression.expr import Expr
        from hdict.expression.step.step import AbsStep

        match other:
            case AbsStep() | hdict() | frozenhdict():
                return Expr(self, other)
            case dict():
                return Expr(self, EDict(other))
            case _:  # pragma: no cover
                return NotImplemented

    def __rrshift__(self, left):
        from hdict import hdict

        if isinstance(left, dict) and not isinstance(left, (hdict, frozenhdict)):
            return frozenhdict(left) >> self
        return NotImplemented  # pragma: no cover

    def __rshift__(self, other):
        # If merging, keep ids.
        from hdict import hdict
        from hdict.expression.step.applyout import ApplyOut
        from hdict.expression.step.cache import cache
        from hdict.content.entry.cached import Cached

        from hdict.content.argument.apply import apply

        if isinstance(other, apply):  # pragma: no cover
            raise Exception(f"Cannot apply without specifying output(s).\n" f"Hint: d >> apply(f)('output_field1', 'output_field2')")
        from hdict.content.argument import AbsArgument

        if isinstance(other, AbsArgument):  # pragma: no cover
            raise Exception(f"Cannot pipe {type(other).__name__} without specifying output.\n" "Hint: d >> {'field name': object}\n" f"Hint: d['field name'] = object")

        from hdict.expression.expr import Expr

        match other:
            case hdict() | frozenhdict():
                dct = other.raw
            case ApplyOut(nested, out):
                dct = {out: nested}
            case cache(storage=storage, fields=fields):
                if not fields:
                    fields = (k for k, v in self.raw.items() if not v.isevaluated)
                dct = {k: Cached(self.ids[k], storage, self.raw[k]) for k in fields}
            case dict():
                dct = other
            case Expr():
                return Expr(self, other).solve()
            case _:  # pragma: no cover
                return NotImplemented
        return frozenhdict(dct, _previous=self)

    def __getitem__(self, item):  # pragma: no cover
        return self.data[item].value

    def __getattr__(self, item):  # pragma: no cover
        if item in self.data:
            return self.data[item].value
        return self.__getattribute__(item)

    def __setitem__(self, key: str, value):  # pragma: no cover
        print(value)
        raise Exception(f"Cannot set an entry ({key}) of a frozen dict.")

    def __delitem__(self, key):  # pragma: no cover
        raise Exception(f"Cannot delete an entry ({key}) from a frozen dict.")

    @staticmethod
    def fromdict(dictionary, ids):
        """Build a frozenidict from values and pre-defined ids

        >>> from hdict import hdict, value
        >>> hdict.fromdict({"x": value(5, hosh="0123456789012345678901234567890123456789")}, {"x": "0123456789012345678901234567890123456789"}).show(colored=False)
        {
            x: 5,
            _id: bi5Qdbh-zgA1ZQdxGhxqjaKaQROtxk1VCPRZhMOq,
            _ids: {
                x: 0123456789012345678901234567890123456789
            }
        }
        """
        from hdict.content.value import value
        from hdict.content.entry import AbsEntry

        data = {}
        for k, v in dictionary.items():
            if isinstance(v, AbsEntry):
                if k in ids and ids[k] != v.id:  # pragma: no cover
                    raise Exception(f"Conflicting ids provided for key '{k}': ival.id={v.id}; ids[{k}]={ids[k]}")
                data[k] = v
            else:
                data[k] = value(v, ids[k] if k in ids else None)
        return frozenhdict(data)

    @property
    def evaluated(self):
        if self._evaluated is None:
            for k, val in self.data.items():
                val.evaluate()
            self._evaluated = self
        return self

    def evaluate(self):
        # todo: add flag to inhibit evaluation (i.e., fetching) of cached values; or other solution, e.g. Cache(write_onapply)
        _ = self.evaluated

    @property
    def asdict(self):
        """
        Convert to 'dict', including ids.

        This evaluates all fields.
        HINT: Use 'dict(d)' to convert a 'hdict' to a 'dict' excluding ids.

        >>> from hdict import hdict, value
        >>> d = hdict.fromdict({"x": value(5, hosh="0123456789012345678901234567890123456789")}, {"x": "0123456789012345678901234567890123456789"})
        >>> d.asdict
        {'x': 5, '_id': 'bi5Qdbh-zgA1ZQdxGhxqjaKaQROtxk1VCPRZhMOq', '_ids': {'x': '0123456789012345678901234567890123456789'}}
        """
        if self._asdict is None:
            dic = dict(self.items())
            dic["_id"] = self.id
            dic["_ids"] = self.ids.copy()
            self._asdict = dic
        return self._asdict

    @property
    def asdicts(self):
        """
        Convert to 'dict' recursing into nested frozenhdicts, including ids.

        This evaluates all fields.
        REMINDER: hdict is never nested, frozenhdict is used instead
        HINT: Use 'asdicts_noid' to recursively convert a 'hdict' to a 'dict' excluding ids.

        >>> from hdict import value, hdict
        >>> d = hdict(x=value(5, hosh="0123456789012345678901234567890123456789"))
        >>> e = hdict(d=d)
        >>> e.asdicts
        {'d': {'x': 5, '_id': 'bi5Qdbh-zgA1ZQdxGhxqjaKaQROtxk1VCPRZhMOq', '_ids': {'x': '0123456789012345678901234567890123456789'}}, '_id': 'GGhKhUmGhISaoHevn39hb-pLZEMoAc3KzE6Z0.IH', '_ids': {'d': 'bi5Qdbh-zgA1ZQdxGhxqjaKaQROtxk1VCPRZhMOq'}}
        >>> dict(e) == e
        True
        """
        if self._asdicts is None:
            from hdict import hdict

            dic = {}
            for k, v in self.items():
                dic[k] = v.asdicts if isinstance(v, (hdict, frozenhdict)) else v
            dic["_id"] = self.id
            dic["_ids"] = self.ids.copy()
            self._asdicts = dic
        return self._asdicts

    @property
    def asdicts_noid(self):
        """
        Convert to 'dict' recursing into nested frozenhdicts, excluding ids.

        REMINDER: hdict is never nested, frozenhdict is used instead
        HINT: Use 'asdicts' to recursively convert a 'hdict' 'd' to 'dict' including ids.

        >>> from hdict import value, hdict
        >>> d = hdict(x=value(5, hosh="0123456789012345678901234567890123456789"))
        >>> e = hdict(d=d)
        >>> e.asdicts_noid
        {'d': {'x': 5}}
        >>> dict(e) == e
        True

        """
        if self._asdicts_noid is None:
            from hdict import hdict

            dic = {}
            for k, v in self.items():
                dic[k] = v.asdicts_noid if isinstance(v, (hdict, frozenhdict)) else v
            self._asdicts_noid = dic
        return self._asdicts_noid

    @property
    def asdicts_hoshes_noneval(self):
        from hdict import value
        from hdict.content.entry import AbsEntry

        hoshes = set()
        dic = {}
        for k, val in self.data.items():
            if isinstance(val, AbsEntry):
                hoshes.add(val.hosh)
            if isinstance(val, value):
                v = val.value
                if isinstance(v, frozenhdict):
                    dic[k], subhoshes = v.asdicts_hoshes_noneval
                    hoshes.update(subhoshes)
                else:
                    dic[k] = v
            else:
                dic[k] = val
        hoshes.add(self.hosh)
        dic["_id"] = self.id
        dic["_ids"] = self.ids.copy()
        return dic, hoshes

    def astext(self, colored=True, key_quotes=False):
        r"""Textual representation of a frozenidict object"""
        dicts, hoshes = self.asdicts_hoshes_noneval
        txt = json.dumps(dicts, indent=4, ensure_ascii=False, cls=CustomJSONEncoder)

        # Put colors after json, to avoid escaping ansi codes.  todo: check how HTML behaves here
        for h in hoshes:
            txt = txt.replace(f'"{h.id}"', repr(h)) if colored else txt.replace(f'"{h.id}"', h.id)

        # Remove quotes.
        txt = re.sub(r'(": )"(λ.+?)"(?=,\n)', '": \\2', txt)  # Closure
        txt = re.sub(r'(": )"(·.+?)"(?=,\n)', '": \\2', txt)  # Wrapper
        txt = re.sub(r'(": )"(↑↓ cached at `.+?·)"(?=,\n)', '": \\2', txt)  # cache
        if not key_quotes:
            txt = re.sub(r'(?<!: )"([\-a-zA-Z0-9_ ]+?)"(?=: )', "\\1", txt)  # keys

        return txt

    def show(self, colored=True, key_quotes=False):
        r"""Print textual representation of a frozenidict object"""
        print(self.astext(colored, key_quotes))

    def copy(self):  # pragma: no cover
        raise Exception("A FrozenIdict doesn't need copies")

    @property
    def unfrozen(self):
        from hdict import hdict

        return hdict(_frozen=self)

    # def entries(self, evaluate=True):
    #     """Iterator over all items"""
    #     yield from self.items(evaluate)
    #     # yield from self.metaitems(evaluate)

    def keys(self):
        """Generator of keys which don't start with '_'"
        >>> from hdict import hdict
        >>> list(hdict(x=3, y=5).keys())
        ['x', 'y']
        >>> list(hdict(x=3, y=5).values())
        [3, 5]
        """
        # return (k for k in self.data if not k.startswith("_"))
        return self.data.keys()

    def values(self, evaluate=True):
        """Generator of field values (keys that don't start with '_')"""
        return ((v.value if evaluate else v) for k, v in self.data.items())
        # return ((v.value if evaluate else v) for k, v in self.data.items() if not k.startswith("_"))

    def items(self, evaluate=True):
        """Generator over field-value pairs"""
        for k, val in self.data.items():
            # if not k.startswith("_"):
            yield k, (val.value if evaluate else val)

    def save(self, storage: dict):
        """
        Store an entire frozenidict
        """
        from hdict.persistence.stored import Stored

        data = {self.id: self.ids}
        for field, fid in self.ids.items():
            value = self[field]
            if field.endswith("_"):
                raise Exception(f"Not implemented for mirror fields")
                # todo:  confirm existence of the counterpart
                # data[kindid(fid)] = str(type(value))
            elif isinstance(value, frozenhdict):
                value.save(storage)
            else:
                data[fid] = Stored(value)
        # todo:  check if frozenhdict is being stored by mistake
        # todo:  attribute/method as subfield:
        #       apply(f, _.df.x)            SubField(name="df", attribute="x")
        #       apply(_.df.drop, "col1")    SubField(name="df", attribute="drop")
        #
        storage.update(data)

    @staticmethod
    def load(id, storage: dict, lazy=True) -> Union["frozenhdict", None]:
        """
        Fetch an entire frozenidict
        """
        if len(id) != 40:  # pragma: no cover
            raise Exception(f"id should have lenght of 40, not {len(id)}")
        return frozenhdict.fetch(id, storage, lazy, ishdict=True)

    @staticmethod
    def fetch(id: str, storage: dict, lazy=True, ishdict=False) -> Union["frozenhdict", None]:
        """
        Fetch a single entry

        When cache is a list, traverse it from the end (right item to the left item).
        """
        from hdict.content.entry.cached import Cached
        from hdict.persistence.stored import Stored

        if id not in storage:
            return None
        obj = storage[id]
        if isinstance(obj, dict):
            ishdict = True  # Set to True, because now we have a nested frozenhdict
        elif ishdict or not isinstance(obj, Stored):  # pragma: no cover
            raise Exception(f"Wrong content for hdict expected under id {id}: {type(obj)}.")

        if ishdict:
            ids = obj
            data = {}
            mirrored = set()
            for field, fid in ids.items():
                if field.endswith("_"):
                    # TODO:  2023-06-23
                    #  -
                    raise Exception(f"Not implemented for mirror fields")
                    # mirrored.add(field[:-1])
                    continue
                if lazy:
                    data[field] = Cached(fid, storage)
                else:
                    obj = frozenhdict.fetch(fid, storage, lazy=False, ishdict=False)
                    if obj is None:  # pragma: no cover
                        print(storage.keys())
                        raise Exception(f"Incomplete hdict: id '{id}' not found in the provided cache.")
                    data[field] = obj
            # for field in mirrored:
            #     obj = data[field]
            #     kind = obj.kind if isinstance(obj, Cached) else getkind(storage, obj.hosh)
            #     data[field + "_"] = handle_mirror(field, data, ids[field], kind)
            return frozenhdict.fromdict(data, ids)

        return obj.content

    @property
    def asdf(self):
        """
        Represent hdict as a DataFrame if possible
        """
        from pandas import DataFrame

        data = dict(self)
        index = data.pop("index")
        return DataFrame(data, index=index)

    @staticmethod
    def fromfile(name, fields=None, format="df", named=None, hide_types=True):
        r"""
        Input format is defined by file extension: .arff, .csv
        """
        from hdict.data.aux_frozendict import handle_format

        df, name = file2df(name, hide_types, True)
        return handle_format(format, fields, df, named and name)

    @staticmethod
    def fromtext(text: str, fields=None, format="df", named=None):
        r"""
        Input format is defined by file extension: .arff, .csv
        """
        from hdict import frozenhdict

        if text.startswith("@"):
            name = "<Unnamed>"
            with StringIO() as f:
                f.write(text)
                text = f.getvalue()
                df = loads(text)
                for line in isplit(text, "\n"):
                    if line[:9].upper() == "@RELATION":
                        name = line[9:].strip()
                        break
        else:
            from testfixtures import TempDirectory

            with TempDirectory() as tmp:
                tmp.write(
                    "temp.csv",
                    text.encode(),
                )
                return frozenhdict.fromfile(tmp.path + "/temp.csv", fields, format, named)

        from hdict.data.aux_frozendict import handle_format

        return handle_format(format, fields, df, named and name)

    def __eq__(self, other):
        if isinstance(other, dict):
            if "_id" in other:
                return self.id == other["_id"]
            if list(self.keys()) != list(other.keys()):
                return False
            from hdict import hdict

            if isinstance(other, (frozenhdict, hdict)):
                return self.id == other.id
            return dict(self) == other
        return NotImplemented  # pragma: no cover

    def __ne__(self, other):
        return not (self == other)

    def __repr__(self):
        return self.astext()

    def __str__(self):
        return stringfy(self.data)

    def __iter__(self):
        for k in self.data:
            yield k

    def __hash__(self):
        return hash(self.hosh)

    def __bool__(self):
        return bool(self.data)

    # def __reduce__(self):
    # dic = self.data.copy()
    # del dic["_id"]
    # del dic["_ids"]
    # return self.__class__, (dic,)



#  Copyright (c) 2023. Davi Pereira dos Santos
#  This file is part of the hdict project.
#  Please respect the license - more about this in the section (*) below.
#
#  hdict is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  hdict is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with hdict.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is illegal and it is unethical regarding the effort and
#  time spent here.
#
from hosh import Hosh

from hdict.content.argument import AbsBaseArgument
from hdict.content.entry import AbsEntry
from hdict.content.aux_value import v2hosh


class value(AbsBaseArgument, AbsEntry):
    """
    Wrapper for any Python object except AbsAny instances

    >>> x = 5
    >>> from hdict.content.value import value
    >>> v = value(x, "1234567890123456789012345678901234567890")
    >>> v
    5
    >>> v.hosh.id
    '1234567890123456789012345678901234567890'

    """

    isevaluated = True

    def __init__(self, val: object, hosh: Hosh | str = None, hdict=None):
        """

        Args:
            val:
            hosh:
            hdict:  optional reference to the object if it has a hdict counterpart (e.g.: pandas DF)
        """
        from hdict.abs import AbsAny

        if isinstance(val, AbsAny):  # pragma: no cover
            raise Exception(f"Cannot handle objects of type '{type(val).__name__}' as raw values for hdict.")
        self.value = self._value = val
        if isinstance(hosh, str):
            hosh = Hosh.fromid(hosh)
        self.hosh = v2hosh(self.value) if hosh is None else hosh
        self.hdict = hdict

    def __repr__(self):
        return repr(self.value)



#  Copyright (c) 2023. Davi Pereira dos Santos
#  This file is part of the hdict project.
#  Please respect the license - more about this in the section (*) below.
#
#  hdict is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  hdict is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with hdict.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is illegal and it is unethical regarding the effort and
#  time spent here.
#
from dataclasses import dataclass
from typing import Callable

from hosh import Hosh

from hdict.abs import AbsAny


@dataclass
class Unevaluated:
    pass


class AbsEntry(AbsAny):
    """
    hdict final entry at internal level: value*, SubValue, Closure

    *value also inherits AbsBaseArgument because 'value' objects have a meaning outside of hdict (external level)
    """

    value: object | Callable  # REMINDER: 'callable' is here for appliable contents, like storing a raw lambda
    hosh: Hosh
    _value = Unevaluated()

    @property
    def id(self):  # pragma: no cover
        return self.hosh.id

    def evaluate(self):
        _ = self.value

    @property
    def isevaluated(self):
        """
        >>> from hdict import apply, frozenhdict
        >>> from hdict.content.entry.closure import Closure
        >>> Closure(apply(lambda x, y: x + y), {"x": 3, "y": 5}, [], frozenhdict).isevaluated
        False
        """
        return not isinstance(self._value, Unevaluated)



from __future__ import annotations

from hdict.content.argument import AbsBaseArgument


def handle_arg(key, val, data, discarded_defaults, out, torepr, previous):
    """Return handled arg and value of pseudocircular entry"""
    from hdict.content.argument.default import default
    from hdict.content.value import value
    from hdict.data.aux_frozendict import handle_item
    from hdict import field
    from hdict.data.aux_frozendict import MissingFieldException

    from hdict.content.argument.entry import entry

    if key == "_":
        return handle_item(str(key), previous, data, previous)
    match val:
        case default(value=v):
            if key in data:
                arg = data[key]
                # REMINDER: `discarded_defaults` exists only for __repr__ purposes
                discarded_defaults.add(key)
            else:
                arg = value(v)
            if key in out:
                torepr[key] = arg
                return arg
        case field(name=name) if name in out:  # Override case of handle_item if pseudocircular reference.
            if name not in data:  # pragma: no cover
                raise MissingFieldException(f"Missing pseudocircular field `{name}`")
            arg = handle_item(str(key), data[name], data, previous)  # key passed for no purpose here AFAIR
            torepr[key] = arg
            return arg
        case entry(name=name):
            from hdict.content.entry.wrapper import Wrapper

            if name not in data:  # pragma: no cover
                raise MissingFieldException(f"Missing pseudocircular field `{name}`")
            content = handle_item(str(key), data[name], data, previous)
            arg = Wrapper(content)
            torepr[f"·{name}"] = arg
            return arg
        case AbsBaseArgument():
            arg = handle_item(str(key), val, data, previous)
        case _:  # pragma: no cover
            raise Exception(f"Cannot handle argument of type `{type(val).__name__}`")
    torepr[key] = val
    return arg



#  Copyright (c) 2023. Davi Pereira dos Santos
#  This file is part of the hdict project.
#  Please respect the license - more about this in the section (*) below.
#
#  hdict is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  hdict is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with hdict.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is illegal and it is unethical regarding the effort and
#  time spent here.
#
from hdict.content.argument import AbsBaseArgument

# todo: accept `apply(lambda _: _.x**2)` as shortcut for `apply(lambda d: d.x**2, _)`  ?
class Self_(AbsBaseArgument):
    """
    >>> from hdict import _, apply
    >>> d = _ >> {"x": 5} >> dict(y=7)
    >>> (d >> apply(lambda a: d.x, _).X).evaluated.show(colored=False)
    {
        x: 5,
        y: 7,
        X: 5,
        _id: I.3UGqgTnpbVa5UdEJGdEvGAX53KEr30qdpnfccX,
        _ids: {
            x: ecvgo-CBPi7wRWIxNzuo1HgHQCbdvR058xi6zmr2,
            y: eJCW9jGsdZTD6-AD9opKwjPIOWZ4R.T0CG2kdyzf,
            X: yVmnByh3DGnZwkav2cyJLm24fxJ6x7pjyQFWovy0
        }
    }
    >>> (d >> apply(lambda _: d.x).X).evaluated.show(colored=False)
    {
        x: 5,
        y: 7,
        X: 5,
        _id: nSe4f8qCC8JFPqvV5MnmpmR1UpdXzIwhkta2DNU3,
        _ids: {
            x: ecvgo-CBPi7wRWIxNzuo1HgHQCbdvR058xi6zmr2,
            y: eJCW9jGsdZTD6-AD9opKwjPIOWZ4R.T0CG2kdyzf,
            X: G8MwJEvu5ZxQiWb9U.roscMuvGuPnvSAssWAM4f9
        }
    }
    >>> type(_), type(+_), type(d)
    (<class 'hdict.Self'>, <class 'hdict.Empty'>, <class 'hdict.hdict'>)
    >>> d.show(colored=False)
    {
        x: 5,
        y: 7,
        _id: A0G3Y7KNMLihDvpSJ3tB.zxshc6u1CbbiiYjCAAA,
        _ids: {
            x: ecvgo-CBPi7wRWIxNzuo1HgHQCbdvR058xi6zmr2,
            y: eJCW9jGsdZTD6-AD9opKwjPIOWZ4R.T0CG2kdyzf
        }
    }
    >>> str(_ * {})
    '⦑{} » {}⦒'
    """

    def __getattr__(self, item):
        from hdict.content.argument.field import field

        return field(item)

    def __getitem__(self, item):
        from hdict.content.argument.field import field

        return field(item)

    def __call__(self, *args, **kwargs):
        from hdict.content.argument.sample import sample

        return sample(*args)

    def __rshift__(self, other):
        from hdict.data.frozenhdict import frozenhdict

        res = frozenhdict().__rshift__(other)
        if res is NotImplemented:  # pragma: no cover
            return res
        if isinstance(res, frozenhdict):
            res = res.unfrozen
        return res

    def __mul__(self, other):
        from hdict.data.frozenhdict import frozenhdict

        res = frozenhdict().__mul__(other)
        if res is NotImplemented:  # pragma: no cover
            return res
        return res.unfrozen

    def __pos__(self):
        from hdict import Ø

        return Ø



from itertools import chain
from typing import Dict
from typing import TypeVar

from hosh import ø

from hdict import hdict, value, frozenhdict, Self
from hdict.abs import AbsAny
from hdict.content.argument.apply import apply
from hdict.content.argument.field import field
from hdict.content.argument.sample import sample
from hdict.content.entry import AbsEntry
from hdict.content.entry.closure import Closure
from hdict.dataset.dataset import df2Xy
from hdict.expression.step.applyout import ApplyOut

VT = TypeVar("VT")


class MissingFieldException(Exception):
    pass


def handle_items(*datas: [Dict[str, object]], previous: frozenhdict):
    result = {} if previous is None else previous.raw.copy()
    result__mirror_fields = {}
    for key, item in chain(*(data.items() for data in datas)):
        entry = handle_item(key, item, result, previous)
        if isinstance(key, str) and key.endswith("_"):
            if isinstance(entry, value):
                result__mirror_fields[f"{key[:-1]}"] = value(entry.hdict, entry.hdict.hosh)
            else:
                raise Exception(f"lazy mirror?")  # todo:
        if isinstance(entry, dict):
            result.update(entry)
        else:
            result[key] = entry
    result.update(result__mirror_fields)
    return result


def handle_item(key, item, result, previous):
    from hdict.content.argument.entry import entry

    match item:
        case AbsEntry():
            res = item
        case field(name=name):
            if name not in result:  # pragma: no cover
                raise MissingFieldException(f"Missing field `{name}`")
            res = handle_item(name, result[name], result, previous)
        case entry(name=name):
            from hdict.content.entry.wrapper import Wrapper

            if name not in result:  # pragma: no cover
                raise MissingFieldException(f"Missing entry `{name}`")
            res = Wrapper(handle_item(name, result[name], result, previous))
        case apply():
            res = item.enclosure(result, key, previous)
        case sample():  # pragma: no cover
            raise Exception(f"Unsampled variable or argument `{key}`")
        case frozenhdict():
            res = value(item, item.hosh)
        case hdict():
            res = value(item.frozen, item.hosh)
        case ApplyOut():  # pragma: no cover
            raise Exception("Cannot assign output through both apply and dict-key: '>> {out: apply(...)(out)}'.")
        case Self():
            if previous is None:
                raise Exception(f"Cannot reference self in a new hdict. `_` is intended to point to a hdict before application.")
            res = handle_item(key, previous, result, None)
        case AbsAny():  # pragma: no cover
            raise Exception(f"Cannot handle instance of type '{type(item).__name__}'.")
        case _ if str(type(item)) == "<class 'pandas.core.frame.DataFrame'>":
            from hdict.dataset.pandas_handling import explode_df

            # todo: check if this is the best default way of handling DFs.
            df = item.copy(deep=False)
            d = explode_df(df)
            df.__dict__["ashdict"] = d
            res = value(df, hosh=d.hosh.rev, hdict=d)
        case _:
            res = value(item)

    if isinstance(key, tuple):
        return handle_multioutput(key, res, result, previous)
    elif not isinstance(key, str):  # pragma: no cover
        raise Exception(f"Invalid type for input field specification: {type(key).__name__}")
    # elif key.startswith("_"):  # pragma: no cover     # reminder: _* allowed here due to parameter name in getattr(__o)
    #     raise Exception(f"Field names cannot start with '_': {key}")

    return res


def handle_identity(data):
    hosh = ø
    ids, later = {}, {}
    for k, v in data.items():
        # Handle meta. mirror, and field ids differently.
        if k.startswith("_"):  # pragma: no cover
            raise Exception("Custom metafields are not allowed:", k)
            # self.mhosh += self.data[k].hosh * k.encode()                # self.mids[k] = self.data[k].hosh.id
        elif k.endswith("_"):
            # mirrorfield, e.g.: 'df_' is a mirror/derived from 'df'
            later[k] = v.id
        else:
            ids[k] = v.hosh.id
        hosh += v.hosh * k.encode()
        # todo:  PAPER REMINDER: state in the paper that identifiers are not strings. they are a special type that never appears as a value.
        #   I.e., hash(identifier) must be different from hash(value), for all identifiers and values.
        #   E.g.: hash(field name X) != hash(string "X")
        #   In this impementation, the difference is always* true because values are always pickled (they are never hashed as strings), while identifiers are just str.encoded().
        #   * → probabilistically
    ids.update(later)
    return hosh, ids


def handle_multioutput(field_names: tuple, entry: AbsEntry | apply, previous_result, previous):
    """Fields and hoshes are assigned to each output according to the alphabetical order of the original keys.

    >>> from hdict import field, value
    >>> d = {"a": field("b"), "b": field("c"), "c": 5}
    >>> d
    {'a': field(b), 'b': field(c), 'c': 5}
    >>> handle_multioutput(("x","y"), value([0, 1]), d, None)
    {'x': 0, 'y': 1}
    >>> handle_multioutput(("x","y"), value({1: "a", 0: "b"}), d, None)
    {'x': 'b', 'y': 'a'}
    """
    from hdict import value
    from hdict.content.entry.subvalue import SubValue

    data = {}
    match entry:
        case value(value=list() as lst):
            if len(field_names) != len(lst):  # pragma: no cover
                raise Exception(f"Number of output fields ('{len(field_names)}') should match number of list elements ('{len(lst)}').")
            for field_name, val in zip(field_names, lst):
                data[field_name] = handle_item(field_name, val, previous_result, previous)
        case value(value=dict() as dct):
            if len(field_names) != len(dct):  # pragma: no cover
                raise Exception(f"Number of output fields ('{len(field_names)}') should match number of dict entries ('{len(dct)}').")
            for field_name, (_, val) in zip(field_names, sorted(dct.items())):
                data[field_name] = handle_item(field_name, val, previous_result, previous)
        case AbsEntry() | apply():
            keys = []  # For repr().
            parent = Closure(entry, previous_result, keys, previous) if isinstance(entry, apply) else entry
            n = len(field_names)
            for key, i, source in loop_field_names(field_names):
                if key is not None:
                    keys.append(key)
                    data[key] = SubValue(parent, i, n, source)
        case _:  # pragma: no cover
            raise Exception(f"Cannot handle multioutput for key '{field_names}' and type '{type(entry).__name__}'.")
    return data


def loop_field_names(field_names):
    if all(isinstance(x, tuple) for x in field_names):
        source_target = sorted((sour, targ) for targ, sour in field_names)
        for i, sour_targ in enumerate(source_target):
            if len(sour_targ) != 2:  # pragma: no cover
                raise Exception(f"Output tuples should be string pairs 'target=source', not a sequence of length '{len(sour_targ)}'.", sour_targ)
            source, target = sour_targ
            yield target, i, source
    elif any(isinstance(x, tuple) for x in field_names):  # pragma: no cover
        raise Exception(f"Cannot mix translated and non translated outputs.", field_names)
    else:
        for i, field_name in enumerate(field_names):
            yield field_name, i, None


# def handle_mirror(k, data, id, kind, previous):  # object | Cached
#     from hdict.content.entry.cached import Cached
#
#     if not isinstance(data[k], (Cached, frozenhdict)):  # pragma: no cover
#         raise Exception(f"Cannot handle fetched object of type `{type(data[k])}`")
#     match kind:
#         case "<class 'pandas.core.frame.DataFrame'>":
#             f = lambda **kwargs: kwargs[k].asdf
#             return apply(f, fhosh=ø, **{k: field(k)}).enclosure(data, k, previous)
#         case None:  # pragma: no cover
#             pass
#         case _:  # pragma: no cover
#             raise Exception(f"Unknown mirror field kind `{kind}`.")


def handle_format(format, fields, df, name):
    """
    >>> from hdict import hdict
    >>> df = hdict(index=[1,2], X=[3,4], y=[5,6]).asdf
    >>> handle_format("Xy", ["X", "y"], df, True).show(colored=False)
    {
        X: "‹{'X': {1: 3, 2: 4}}›",
        y: "‹[0 1]›",
        name: true,
        _id: rd.NmqV3FPKd3o5c7XAcPcJ6GFkK7O5XSAZR7JkV,
        _ids: {
            X: OMVTYSyh4Tl7V-eDNkbZ.1j79E5eLBOPLtXB.87w,
            y: cm5S71YBRAlVWV5Yn9pXHzsacyZuEH4ZFDNAw9nu,
            name: oK8X-7eG1Qp1WH7v6fokBDrQPdngKn.h86tlEnx4
        }
    }
    >>> handle_format("df", ["X", "y"], df, True).show(colored=False)
    {
        df: "‹{'X': {1: 3, 2: 4}, 'y': {1: 5, 2: 6}}›",
        name: true,
        _id: VWDHscqLlIW-Fu6bOAjrEBHzUehukqBOxK3RM4m9,
        _ids: {
            df: wTtDSUgXcZQVqx-CBZI-L9h4zVjAazPRGQOUNza8,
            name: oK8X-7eG1Qp1WH7v6fokBDrQPdngKn.h86tlEnx4
        }
    }
    >>> handle_format("Xy", None, df, True).show(colored=False)
    {
        X: "‹{'X': {1: 3, 2: 4}}›",
        y: "‹[0 1]›",
        name: true,
        _id: rd.NmqV3FPKd3o5c7XAcPcJ6GFkK7O5XSAZR7JkV,
        _ids: {
            X: OMVTYSyh4Tl7V-eDNkbZ.1j79E5eLBOPLtXB.87w,
            y: cm5S71YBRAlVWV5Yn9pXHzsacyZuEH4ZFDNAw9nu,
            name: oK8X-7eG1Qp1WH7v6fokBDrQPdngKn.h86tlEnx4
        }
    }
    """
    if fields and not isinstance(fields, list):  # pragma: no cover
        raise Exception(f"`fields` must be a list.")
    if fields is None:
        fields = ["df"]
    if format == "df":
        if fields == ["X", "y"]:
            fields = ["df"]
        if len(fields) != 1:  # pragma: no cover
            raise Exception(f"Wrong number of fields {len(fields)}. Expected: 1.", fields)
        d = frozenhdict({fields[0]: df})
    elif format == "Xy":
        if fields == ["df"]:
            fields = ["X", "y"]
        if len(fields) != 2:  # pragma: no cover
            raise Exception(f"Wrong number of fields {len(fields)}. Expected: 2.", fields)
        dic = df2Xy(df=df)
        d = frozenhdict({fields[0]: dic["X"], fields[1]: dic["y"]})
    else:  # pragma: no cover
        raise Exception(f"Unknown {format=}.")

    if name:
        d >>= {"name": name}
    return d



#  Copyright (c) 2023. Davi Pereira dos Santos
#  This file is part of the hdict project.
#  Please respect the license - more about this in the section (*) below.
#
#  hdict is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  hdict is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with hdict.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is illegal and it is unethical regarding the effort and
#  time spent here.
#


def explode_df(df):
    """
    >>> from pandas import DataFrame
    >>> from hdict import hdict, cache
    >>> df = DataFrame({"x": [1,2,3], "y": [5,6,7]}, index=["a", "b", "c"])
    >>> d = hdict(df_=df)
    >>> d.show(colored=False)
    {
        df_: "‹{'x': {'a': 1, 'b': 2, 'c': 3}, 'y': {'a': 5, 'b': 6, 'c': 7}}›",
        df: {
            index: "‹{'a': 'a', 'b': 'b', 'c': 'c'}›",
            x: "‹{'a': 1, 'b': 2, 'c': 3}›",
            y: "‹{'a': 5, 'b': 6, 'c': 7}›",
            _id: efFl2-CCjnStQSpp8QmwkK-7ANKh8bNctOap9X9m,
            _ids: {
                index: HdOKL6NHC9ApFmGygZ54A9f265R6TxhGjBRwdO9r,
                x: n7fkFYWuvJ.MTp7asUu8489mQMwoZiL.5.or1EDt,
                y: N9Of5idOlZFH25hFb9IcjOwJrxt2.tXddWKuYq94
            }
        },
        _id: symsSXy-oMJrIpAisV7aHbn.fZ9yfjfrExxIq6HS,
        _ids: {
            df: efFl2-CCjnStQSpp8QmwkK-7ANKh8bNctOap9X9m,
            df_: e.VIUfoRxV4B6aoIZeJ8AlDwmq7IZNPDNPFI72WZ
        }
    }
    >>> d.df_
       x  y
    a  1  5
    b  2  6
    c  3  7
    >>> c = {}
    >>> d >>= cache(c)
    >>> d.show(colored=False)
    {
        df_: "‹{'x': {'a': 1, 'b': 2, 'c': 3}, 'y': {'a': 5, 'b': 6, 'c': 7}}›",
        df: {
            index: "‹{'a': 'a', 'b': 'b', 'c': 'c'}›",
            x: "‹{'a': 1, 'b': 2, 'c': 3}›",
            y: "‹{'a': 5, 'b': 6, 'c': 7}›",
            _id: efFl2-CCjnStQSpp8QmwkK-7ANKh8bNctOap9X9m,
            _ids: {
                index: HdOKL6NHC9ApFmGygZ54A9f265R6TxhGjBRwdO9r,
                x: n7fkFYWuvJ.MTp7asUu8489mQMwoZiL.5.or1EDt,
                y: N9Of5idOlZFH25hFb9IcjOwJrxt2.tXddWKuYq94
            }
        },
        _id: symsSXy-oMJrIpAisV7aHbn.fZ9yfjfrExxIq6HS,
        _ids: {
            df: efFl2-CCjnStQSpp8QmwkK-7ANKh8bNctOap9X9m,
            df_: e.VIUfoRxV4B6aoIZeJ8AlDwmq7IZNPDNPFI72WZ
        }
    }
    >>> d.df.show(colored=False)
    {
        index: "‹{'a': 'a', 'b': 'b', 'c': 'c'}›",
        x: "‹{'a': 1, 'b': 2, 'c': 3}›",
        y: "‹{'a': 5, 'b': 6, 'c': 7}›",
        _id: efFl2-CCjnStQSpp8QmwkK-7ANKh8bNctOap9X9m,
        _ids: {
            index: HdOKL6NHC9ApFmGygZ54A9f265R6TxhGjBRwdO9r,
            x: n7fkFYWuvJ.MTp7asUu8489mQMwoZiL.5.or1EDt,
            y: N9Of5idOlZFH25hFb9IcjOwJrxt2.tXddWKuYq94
        }
    }
    """
    from hdict.data.frozenhdict import frozenhdict

    dic = {"index": df.index.to_series()}
    for col in df:
        dic[str(col)] = df[col]
    d = frozenhdict(dic)
    return d


def file2df(filename, hide_types=True, return_name=True, transpose=False, index=False):
    from hdict.dataset.dataset import load

    if filename.endswith(".arff"):
        relation = None
        with open(filename) as f:
            for line in f:
                if line[:9].upper() == "@RELATION":
                    relation = line[9:-1]
                    break
        with open(filename) as f:
            df = load(f)

        if index or transpose:
            indexname = df.columns[0]
            df.set_index(indexname, inplace=True)
            df.index.name = indexname
        if transpose:
            df = df.T
            if index:
                df.index.rename(indexname, inplace=True)

        if hide_types:
            df.rename(columns={k: k.split("@")[0] for k in df.columns}, inplace=True)
        if return_name:
            return df, relation or filename
        else:
            return df
    elif filename.endswith(".csv"):
        from pandas import read_csv

        df = read_csv(filename)

        if index or transpose:
            indexname = df.columns[0]
            df.set_index(indexname, inplace=True)
            df.index.name = indexname
        if transpose:
            df = df.T
            if index:
                df.index.rename(indexname, inplace=True)

        if return_name:
            return df, filename
        else:
            return df
    else:  # pragma: no cover
        raise Exception(f"Unknown extension {filename.split('.')[-1]}.")



#  Copyright (c) 2023. Davi Pereira dos Santos
#  This file is part of the hdict project.
#  Please respect the license - more about this in the section (*) below.
#
#  hdict is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  hdict is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with hdict.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is illegal and it is unethical regarding the effort and
#  time spent here.
#

from hdict.expression.step.step import AbsStep


class cache(AbsStep):
    """
    >>> from hdict import _, hdict, cache, apply
    >>> storage = {}
    >>> d = hdict(x=3, y=5) >> apply(lambda x, y: x / y).z
    >>> d.show(colored=False)
    {
        x: 3,
        y: 5,
        z: λ(x y),
        _id: IIh20mb2BGUx9XMCDwzTr4y7v-JaGin3eEHvx692,
        _ids: {
            x: KGWjj0iyLAn1RG6RTGtsGE3omZraJM6xO.kvG5pr,
            y: ecvgo-CBPi7wRWIxNzuo1HgHQCbdvR058xi6zmr2,
            z: GC-BMZNVRagHxgUynUadKUYA1kZ8GzcUj6OKWstQ
        }
    }
    >>> d >>= cache(storage)
    >>> d.show(colored=False)
    {
        x: 3,
        y: 5,
        z: ↑↓ cached at `dict`·,
        _id: IIh20mb2BGUx9XMCDwzTr4y7v-JaGin3eEHvx692,
        _ids: {
            x: KGWjj0iyLAn1RG6RTGtsGE3omZraJM6xO.kvG5pr,
            y: ecvgo-CBPi7wRWIxNzuo1HgHQCbdvR058xi6zmr2,
            z: GC-BMZNVRagHxgUynUadKUYA1kZ8GzcUj6OKWstQ
        }
    }
    >>> d.evaluate()
    >>> d.show(colored=False)
    {
        x: 3,
        y: 5,
        z: 0.6,
        _id: IIh20mb2BGUx9XMCDwzTr4y7v-JaGin3eEHvx692,
        _ids: {
            x: KGWjj0iyLAn1RG6RTGtsGE3omZraJM6xO.kvG5pr,
            y: ecvgo-CBPi7wRWIxNzuo1HgHQCbdvR058xi6zmr2,
            z: GC-BMZNVRagHxgUynUadKUYA1kZ8GzcUj6OKWstQ
        }
    }
    """

    def __init__(self, storage: dict, *fields):
        self.storage = storage
        self.fields = fields

    def __repr__(self):
        return f"↑↓`{type(self.storage).__name__}`"



#  Copyright (c) 2023. Davi Pereira dos Santos
#  This file is part of the hdict project.
#  Please respect the license - more about this in the section (*) below.
#
#  hdict is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  hdict is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with hdict.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is illegal and it is unethical regarding the effort and
#  time spent here.
#

from hdict.data.frozenhdict import frozenhdict


class Empty_(frozenhdict):
    """
    >>> from hdict import Ø
    >>> d = Ø >> {"x": 5} >> dict(y=7)
    >>> type(Ø), type(d)
    (<class 'hdict.Empty'>, <class 'hdict.hdict'>)
    >>> d.show(colored=False)
    {
        x: 5,
        y: 7,
        _id: A0G3Y7KNMLihDvpSJ3tB.zxshc6u1CbbiiYjCAAA,
        _ids: {
            x: ecvgo-CBPi7wRWIxNzuo1HgHQCbdvR058xi6zmr2,
            y: eJCW9jGsdZTD6-AD9opKwjPIOWZ4R.T0CG2kdyzf
        }
    }
    >>> str(Ø * {})
    '⦑{} » {}⦒'
    """

    def __rshift__(self, other):
        res = super().__rshift__(other)
        if res is NotImplemented:  # pragma: no cover
            return res
        if isinstance(res, frozenhdict):
            res = res.unfrozen
        return res

    def __mul__(self, other):
        res = super().__mul__(other)
        if res is NotImplemented:  # pragma: no cover
            return res
        return res.unfrozen



from functools import reduce
from operator import rshift
from random import Random

from hdict.abs import AbsAny
from hdict.expression.step.step import AbsStep


class Expr(AbsStep):
    """
    Expressions enable the creation of pipelines of steps (or nested expressions)

    Pipelines should be sampled or fed a hdict to become a resulting hdict.

    >>> from hdict import hdict
    >>> from hdict.expression.expr import Expr
    >>> e = Expr() >> hdict()
    >>> e.show(colored=False)
    ⦑{
        _id: 0000000000000000000000000000000000000000,
        _ids: {}
    }⦒
    >>> e = Expr() >> dict()
    >>> e.show(colored=False)
    ⦑{}⦒
    """

    def __init__(self, *steps):
        self.steps = steps

    def sample(self, rnd: int | Random = None, solve=True):
        from hdict.content.argument import AbsArgument
        from hdict.expression.step.applyout import ApplyOut
        from hdict import cache, hdict
        from hdict.data.frozenhdict import frozenhdict
        from hdict.expression.step.edict import EDict

        lst = []
        for step in self:
            match step:
                case AbsArgument() | ApplyOut():
                    lst.append(step.sample(rnd))
                case cache():
                    lst.append(step)
                case EDict():
                    dct = step.dct.copy()
                    for k, v in dct.items():
                        if isinstance(v, AbsArgument):
                            dct[k] = dct[k].sample(rnd)
                    lst.append(EDict(dct))
                case frozenhdict() | hdict():
                    lst.append(step)
                case AbsAny():  # pragma: no cover
                    raise Exception(f"{type(step)}")
                case x:  # pragma: no cover
                    raise Exception(f"{type(x)}")
        expr = Expr.fromiter(lst)
        return expr.solve() if solve else expr

    def __invert__(self):
        return self.sample()

    @staticmethod
    def fromiter(lst):
        new = Expr()
        new.steps = lst
        return new

    def solve(self):
        from hdict.expression.step.edict import EDict

        gen = (step.dct if isinstance(step, EDict) else step for step in self)
        return reduce(rshift, gen)

    @property
    def unfrozen(self):
        from hdict import frozenhdict

        gen = (step.unfrozen if isinstance(step, frozenhdict) else step for step in self)
        return Expr.fromiter(gen)

    def __iter__(self):
        for step in self.steps:
            if isinstance(step, Expr):
                yield from step
            else:
                yield step

    def show(self, colored=True, key_quotes=False):
        r"""Print textual representation of an expression"""
        print(self.astext(colored, key_quotes))

    def astext(self, colored=True, key_quotes=False):
        r"""
        Textual representation of an expression

        >>> p = Expr({"b":2}, {"a":1})
        >>> p
        ⦑{'b': 2} » {'a': 1}⦒
        >>> p.show()
        ⦑{'b': 2} » {'a': 1}⦒
        """
        from hdict.data.frozenhdict import frozenhdict
        from hdict import hdict

        out = []
        for step in self:
            if isinstance(step, (frozenhdict, hdict)):
                out.append(step.astext(colored=colored, key_quotes=key_quotes))
            else:
                out.append(repr(step))
        return "⦑" + " » ".join(out) + "⦒"

    def __repr__(self):
        return "⦑" + " » ".join(repr(step) for step in self) + "⦒"

    def __str__(self):
        out = []
        for step in self:
            out.append(str(step))
        return "⦑" + " » ".join(out) + "⦒"

    def __bool__(self):  # pragma: no cover
        raise Exception(f"Cannot know the bool value of an expression before evaluating it.")
